"""
Extracts the earthquake catalog from IRIS for a given station and time period based on obspy 
"""
# Download the catalog
import os
from os import mkdir
from os.path import exists, dirname
from obspy.clients.fdsn.client import Client
from obspy import UTCDateTime
from obspy.core.event.catalog import read_events
from rfsed.util import save_catalog
savepath=save_catalog()
if not exists(savepath):
    mkdir(savepath)
#Station Information
staname='NE301'
stalat=6.183432
stalon=53.48334
client = Client("IRIS")
starttime = UTCDateTime("2022-02-01")
endtime = UTCDateTime("2022-03-31")
f = "%seq_events.quakeml"%(savepath)
#% Get the catalog and save it
cat = client.get_events(starttime = starttime, endtime = endtime, latitude=stalat, longitude=stalon,\
                        minmagnitude = 6.0, maxmagnitude=10.0, minradius=30, maxradius=90, filename = f)

# Read the file again and plot it
# cat = read_events(f, format="QUAKEML")
# cat.plot()
# %%



# Copyright 2013-2016 Tom Eulenfeld, MIT license
"""
Tests for deconvolve module.
"""
import unittest

import numpy as np
from scipy.signal import get_window, convolve
from rfsed.rf.rfstream import read_rf, rfstats


def test_deconvolve_Lpeak(testcase, stream, *args, **kwargs):
    s1 = stream.copy()
    s1.deconvolve(*args, **kwargs)
    Ltrace = s1.select(component='L')[0]
    st = Ltrace.stats
    onsetL = Ltrace.data.argmax() * st.delta
    onset = st.onset - st.starttime
    msg = ('L component maxium at %.2fs, but should be at %.2fs. '
           'deconvolve args: %s, %s') % (onsetL, onset, args, kwargs)
    testcase.assertLess(abs(onsetL-onset), 0.1, msg=msg)
    msg = 'maximum of L component is %.2f, but should 1'
    max_ = Ltrace.data.max()
    testcase.assertGreater(max_, 0.999, msg=msg % max_)


def test_deconvolve_Qpeak(testcase, stream, *args, **kwargs):
    s1 = stream.copy()
    winsrc = kwargs.pop('winsrc', 'S')
    s1.deconvolve(*args, winsrc=winsrc, source_components='Q', **kwargs)
    Qtrace = s1.select(component='Q')[0]
    st = Qtrace.stats
    onsetQ = Qtrace.data.argmax() * st.delta
    onset = st.onset - st.starttime
    msg = ('Q component maxium at %.2fs, but should be at %.2fs. '
           'deconvolve args: %s, %s') % (onsetQ, onset, args, kwargs)
    testcase.assertLess(abs(onsetQ-onset), 0.01, msg=msg)
    msg = 'maximum of Q component is %.2f, but should 1'
    max_ = Qtrace.data.max()
    testcase.assertGreater(max_, 0.999, msg=msg % max_)


class DeconvolveTestCase(unittest.TestCase):

#    # should be tested in toeplitz module
#    def test_toeplitz_real_sym(self):
#        # set specific seed value such that random numbers are reproducible
#        seed(0)
#        src = random(50) - 0.5
#        rsp = random(50) - 0.5
#        toep = scipy.linalg.toeplitz(src)
#        x = np.dot(scipy.linalg.inv(toep), rsp)  # compare to scipy.linalg
#        x2 = rfsed.rf.deconvolve._toeplitz_real_sym(src, rsp)
#        np.testing.assert_array_almost_equal(x, x2, decimal=3)

    def test_deconvolution_of_stream_Lpeak_position(self):
        stream = read_rf()[:3]
        rfstats(stream)
        stream.filter('bandpass', freqmin=0.4, freqmax=1)
        stream.trim2(5, 95, reftime='starttime')
        stream.rotate('ZNE->LQT')
        # check that maximum in L component is at 0s (at P onset)
        test_deconvolve_Lpeak(self, stream, 'time')
        test_deconvolve_Lpeak(self, stream, 'time', solve_toeplitz='scipy')
        test_deconvolve_Lpeak(self, stream, 'waterlevel')
        test_deconvolve_Lpeak(self, stream, 'iterative')
        try:
            test_deconvolve_Lpeak(self, stream, 'multitaper')
        except ImportError:
            pass
        test_deconvolve_Lpeak(self, stream, 'time', winsrc=(-20, 40, 5))
        test_deconvolve_Lpeak(self, stream, 'waterlevel', winsrc=(-20, 40, 5))
        stream.trim2(5, 70, reftime='starttime')
        test_deconvolve_Lpeak(self, stream, 'time')
        test_deconvolve_Lpeak(self, stream, 'waterlevel')
        test_deconvolve_Lpeak(self, stream, 'time', winsrc=(-20, 40, 5))
        test_deconvolve_Lpeak(self, stream, 'waterlevel', winsrc=(-20, 40, 5))
        test_deconvolve_Lpeak(self, stream, 'time', winsrc=(-5, 18, 5))
        test_deconvolve_Lpeak(self, stream, 'waterlevel', winsrc=(-5, 18, 5))

    def test_deconvolution_of_stream_Qpeak_position(self):
        # S receiver deconvolution
        from pkg_resources import resource_filename
        fname = resource_filename('rfsed', 'rf/example/minimal_example_S.tar.gz')
        stream = read_rf(fname)[:3]
        rfstats(stream, phase='S')
        stream.filter('bandpass', freqmin=0.2, freqmax=0.5)
        stream.trim2(10, 120, reftime='starttime')
        stream.rotate('ZNE->LQT')
        # check that maximum in Q component is at 0s (at S onset)
        test_deconvolve_Qpeak(self, stream, 'time')
        test_deconvolve_Qpeak(self, stream, 'time', solve_toeplitz='scipy')
        test_deconvolve_Qpeak(self, stream, 'waterlevel')
        test_deconvolve_Qpeak(self, stream, 'iterative')
        try:
            test_deconvolve_Qpeak(self, stream, 'multitaper')
        except ImportError:
            pass
        test_deconvolve_Qpeak(self, stream, 'time', winsrc=(-5, 18, 5))
        test_deconvolve_Qpeak(self, stream, 'waterlevel', winsrc=(-5, 18, 5))
        test_deconvolve_Qpeak(self, stream, 'time', winsrc=(-20, 40, 5))
        test_deconvolve_Qpeak(self, stream, 'waterlevel', winsrc=(-20, 40, 5))

    def test_deconvolution_of_convolution(self):
        from rfsed.rf.rfstream import RFStream, RFTrace
        data = np.zeros(400)
        data_src = np.zeros(400)
        hann1 = get_window('hann', 10)
        hann2 = get_window('hann', 50)
        data_src[40:50] = hann1
        data_src[50:60] = -hann1
        data[100:150] = hann2
        data[240:290] = 0.5 * hann2
        data_rsp = convolve(data_src, data, 'full')[50:450]/3.
        stream1 = RFStream([RFTrace(data=data_src), RFTrace(data=data_rsp)])
        for i, tr in enumerate(stream1):
            tr.stats.channel = tr.stats.channel[:2] + 'LQT'[i]
            tr.stats.onset = tr.stats.starttime + 40
        stream2 = stream1.copy()
        stream3 = stream1.copy()
        stream4 = stream1.copy()
        stream1.deconvolve(spiking=10)
        stream2.deconvolve(method='waterlevel', waterlevel=0.1)
        stream3.deconvolve(method='iterative')
        try:
            stream4.deconvolve(method='multitaper')
        except ImportError:
            pass

#        import matplotlib.pyplot as plt
#        plt.subplot(121)
#        plt.plot(data, label='desired')
#        plt.plot(data_src, label='source')
#        plt.plot(data_rsp, label='convolution')
#        plt.plot(stream1[0].data, label='deconv src')
#        plt.plot(stream1[1].data, label='deconv')
#        plt.legend()
#        plt.subplot(122)
#        plt.plot(data)
#        plt.plot(data_src)
#        plt.plot(data_rsp)
#        plt.plot(stream2[0].data)
#        plt.plot(stream2[1].data)
#        plt.show()

        # (shift from middle of source (50) to onset (40)
        peakpos = np.argmax(data) - 10
        self.assertEqual(peakpos, np.argmax(stream1[1].data))
        self.assertEqual(peakpos, np.argmax(stream2[1].data))
        self.assertEqual(peakpos, np.argmax(stream3[1].data))

    def test_deconvolve_custom_function(self):
        def test_func_stream(rsp, src, **kw):
            for r_ in rsp:
                r_.data[:] = 1
                r_.stats.custom = 'X'
            self.assertTrue(hasattr(src, 'stats'))
            return rsp

        stream = read_rf()[:3]
        rfstats(stream)
        stream1 = stream.copy()
        stream1.rf(deconvolve='func', func=test_func_stream)
        for tr in stream1:
            np.testing.assert_equal(tr.data, 1)
            self.assertIn('custom', tr.stats)
        stream1 = stream
        stream1.deconvolve(method='func', func=test_func_stream)
        for tr in stream1:
            np.testing.assert_equal(tr.data, 1)
            self.assertIn('custom', tr.stats)


def suite():
    return unittest.makeSuite(DeconvolveTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



"""
.. module:: ExtractEq
        :synopsis: Extract the data for each earthquake from the local data files
.. moduleauthor:: Stephen Akinremi <s.akinremi@utwente.nl> Islam Fadel <i.e.a.m.fadel@utwente.nl> (October 2023)
"""

from obspy import read, read_inventory, Stream, read_events, UTCDateTime as UTC
import numpy as np
from obspy.core.event.catalog import read_events
from glob import glob
from obspy.geodetics.base import locations2degrees, gps2dist_azimuth
from obspy.geodetics import kilometer2degrees
from obspy.taup import TauPyModel
#------------------------------------------------------------------------------------
def ExtractEq(datapath, filename, catalog, stalat, stalon, Request_window):
        """
        Extract the data for each earthquake from the local data files
        Writes the extracted data to a new file

        :param datapath: Path to the local data files directory
        :type datapath:str
        :param filename: Path to the new file
        :type filename:str
        :param catalog: Path to the catalog file
        :type catalog:str
        :param stalat: Latitude of the station
        :type stalat:float
        :param stalon: Longitude of the station
        :type stalon:float
        :param Request_window: Time window relative to first P arrival
        :type Request_window:list[start, end]
        :return: Saves extracted earthquake data into a new file
        """
        datafiles = sorted(glob("%s*.dat"%(datapath)))
        len_datafiles = len(datafiles)
        firstfile = read(datafiles[0])
        lastfile = read(datafiles[len_datafiles - 1])
        recordstarttime = firstfile[0].stats.starttime 
        recordendtime = lastfile[0].stats.endtime
        # staname=firstfile[0].stats.station
        # print('Start of all data record = ', recordstarttime)
        # print('End of all data record = ', recordendtime)
        #----------------------------------------------
        # Get the requested UTC time window for each earthquake
        catalog = read_events(catalog)
        len_catalog=len(catalog)
        # print(len_catalog, ' events in Catalog')    
        time_window = []
        for i in catalog:
                onsettime = i.origins[0].time
                eqlat = i.origins[0].latitude
                eqlon = i.origins[0].longitude
                eqmag = i.magnitudes[0].mag
                eqdepth = (i.origins[0].depth)/1000
                distm,baz,az  = gps2dist_azimuth(stalat, stalon, eqlat, eqlon)
                distkm = distm/1000
                distdeg = locations2degrees(stalat, stalon, eqlat, eqlon)  
                model = TauPyModel(model="iasp91")   
                traveltime = model.get_travel_times(source_depth_in_km=eqdepth, distance_in_degree=distdeg)
                arrivals=traveltime
                P = traveltime[0].time   
                rayparam = kilometer2degrees(1) * traveltime[0].ray_param_sec_degree
                incidangle = traveltime[0].incident_angle
                eqstart = (onsettime + P) - abs(Request_window[0])
                eqend = (onsettime + P) + abs(Request_window[1])
                time_window.append([eqstart, eqend])
        #------------------------------------------------------------------------------------
        # Read part of the data with chosen Earthquakes and save to a new file
        Datast = Stream()
        for i in datafiles:
                st = read(i)
                filestart=st[0].stats.starttime
                fileend = st[0].stats.endtime
                for t in time_window:
                        if UTC(t[0]) >= filestart and UTC(t[1]) <= fileend:
                                Eqdata=read(i, starttime=UTC(t[0]), endtime = UTC(t[1]))
                                Datast += Eqdata
        Datast.write(filename)  
#--------------------------------------------------------------------------------


"""
.. module:: ExtractEqMultiproc
        :synopsis: Extract the data for each earthquake from the local data files in parallel
.. moduleauthor:: Stephen Akinremi <s.akinremi@utwente.nl> Islam Fadel <i.e.a.m.fadel@utwente.nl> (October 2023)
"""

from obspy import read, read_events, Stream, UTCDateTime as UTC
import numpy as np
from obspy.core.event.catalog import read_events
from obspy.geodetics.base import locations2degrees, gps2dist_azimuth
from obspy.geodetics import kilometer2degrees
from obspy.taup import TauPyModel
import multiprocessing
import itertools

def Add_sta_time_2cat(catalogfile, stalat, stalon, Request_window):
    """
    Add station latitude, longitude and request time window to each event in the catalog

    :param catalogfile: Path to the catalog file
    :type catalogfile: str
    :param stalat: Station latitude
    :type stalat: float
    :param stalon: Station longitude
    :type stalon: float
    :param Request_window: Request time window relative to first P arrival (time before P arrival, time after P arrival)
    :type Request_window: list [start, end]
    :return: Catalog with station latitude, longitude and request time window
    """
    catalog = read_events(catalogfile)
    for i in catalog:
        i.stalat = stalat
        i.stalon = stalon
        i.Request_start = Request_window[0]
        i.Request_end = Request_window[1]
    return catalog
#----------------------------------------------
def Eq_times(catalog):
    """
    Get the earthquake start and end time for the requested time window relative to the 
    first P arrival for each event in the catalog. The theoretical P wave arrival time is calculated using the IASP91 reference model

    :param catalog: Catalog with station latitude, longitude and request time window (output of Add_cat_sta_time)
    :type catalog: obspy.core.event.catalog.Catalog
    :return: Dictionary of Earthquake start and end time for the requested time window relative to the first P arrival
    """
    stalat = catalog.stalat
    stalon = catalog.stalon
    Request_window = [catalog.Request_start, catalog.Request_end]       
    onsettime = catalog.origins[0].time
    eqlat = catalog.origins[0].latitude
    eqlon = catalog.origins[0].longitude
    eqmag = catalog.magnitudes[0].mag
    eqdepth = (catalog.origins[0].depth)/1000
    distm,baz,az  = gps2dist_azimuth(stalat, stalon, eqlat, eqlon)
    distkm = distm/1000
    distdeg = locations2degrees(stalat, stalon, eqlat, eqlon)    
    model = TauPyModel(model="iasp91")   
    arrivals = model.get_travel_times(source_depth_in_km=eqdepth, distance_in_degree=distdeg)
    P = arrivals[0].time   
    rayparam = kilometer2degrees(1) * arrivals[0].ray_param_sec_degree
    incidangle = arrivals[0].incident_angle
    eqstart = (onsettime + P) - abs(Request_window[0])
    eqend = (onsettime + P) + abs(Request_window[1])
    timewindow=[{'eqstart':eqstart, 'eqend':eqend}]
    return timewindow
#----------------------------------------------
# Run Parallel processing
def Get_Eqtimes_multiproc(catalog, nproc):
    """
    Run the Eq_times function in parallel for each event in the catalog

    :param catalog: Catalog with station latitude, longitude and request time window (output of Add_cat_sta_time)
    :type catalog: obspy.core.event.catalog.Catalog
    :param nproc: Number of processors to use
    :type nproc: int
    :return: List of dictionaries of Earthquake start and end time for the requested time window relative to the first P arrival
    """
    nproc = nproc
    pool = multiprocessing.get_context('fork').Pool(nproc)
    timewindow = pool.map(Eq_times, catalog)
    pool.close()
    pool.join()
    return timewindow
#----------------------------------------------
def Get_EqStream_multiproc(inputparams):
    """
    Extract the earthquake data from the local data files for the requested time window relative to the first P arrival

    :param inputparams: List of input parameters (datafiles, timewindow)
    :type inputparams: list
    :return: Stream of earthquake data for the requested time window relative to the first P arrival
    """
    datafiles, timewindow = inputparams
    Datast = Stream()
    Allstream = Stream()
    datafiles = datafiles[0]
    for i in timewindow:
        eqstart=(i[0]['eqstart'])
        eqend=(i[0]['eqend'])
        Eqdata=read(datafiles, starttime=eqstart, endtime = eqend)
        for i in Eqdata:
            Datast.append(i)
    return Datast
# ---------------------------------------------- 
def ExtractEq_Multiproc(datafiles, timewindow, nproc, filename):
    """
    Run the Get_EqStream_multiproc function in parallel for each data file, extracting the earthquake data 
    for the requested time window relative to the first P arrival

    :param datafiles: List of data files
    :type datafiles: list
    :param timewindow: Dictionary of Earthquake start and end time for the requested time window relative to the first P arrival
    :type timewindow: dict
    :param nproc: Number of processors to use
    :type nproc: int
    :param filename: Path to save the extracted earthquake data
    :type filename: str
    :return: Saves the extracted earthquake data to the specified path
    """
    nproc = nproc
    with multiprocessing.Pool(nproc) as pool:
        inputparams=((inputparams, timewindow) for inputparams in itertools.product(datafiles))
        EqStreams=pool.map(Get_EqStream_multiproc, inputparams)   
    pool.close()
    pool.join()
    #----------------------------------------------
    AllStream = Stream()
    for i in range(0,len(EqStreams)):
            AllStream += EqStreams[i]
    AllStream.write(filename)
#----------------------------------------------------------------------------------------------------



# Copyright 2013-2016 Tom Eulenfeld, MIT license
"""
Tests for rfstream module.
"""
import unittest

from obspy import read, read_events
from obspy.core import AttribDict
from obspy.core.util import NamedTemporaryFile
from rfsed.rf import read_rf, RFStream, rfstats
from rfsed.rf.rfstream import (obj2stats, _HEADERS, _STATION_GETTER, _EVENT_GETTER,
                         _FORMATHEADERS)
from rfsed.rf.util import minimal_example_rf, minimal_example_Srf

_HEADERS_TEST_IO = (50.3, -100.2, 400.3,  # station coordinates
                    -20.32, 10., 12.4, 6.5, -40.432,  # event properties
                    20.643,  # onset
                    'rf', 'P', 'Ps',  # type, phase, moveout
                    57.6, 90.1, 10.2, 10.,  # arrival properties
                    10., -20, 150,  # piercing points
                    15.7, 2.5)  # box properties

_HEADERS_NOT_BY_RFSTATS = ('moveout', 'box_pos', 'box_length', 'type')

FORMATS = list(_FORMATHEADERS.keys())

try:
    import obspyh5
except ImportError:
    obspyh5 = None
else:
    FORMATS.append('h5')


def write_test_header(stream):
    for tr in stream:
        st = tr.stats
        for head, val in zip(_HEADERS, _HEADERS_TEST_IO):
            if head in ('onset', 'event_time'):
                val = st.starttime + val
            st[head] = val


def test_io_header(testcase, stream, ignore=()):
    for format in FORMATS:
        stream1 = stream.copy()
        suffix = '.' + format.upper()
        if format == 'sh':
            format = 'q'
            suffix = '.QHD'
        elif format == 'h5':
            for tr in stream1:
                tr.stats.pop('sac', None)
        with NamedTemporaryFile(suffix=suffix) as ft:
            fname = ft.name
            stream1.write(fname, format.upper())
            stream2 = read_rf(fname)
        st1 = stream1[0].stats
        st2 = stream2[0].stats
        for head in _HEADERS:
            if head in st1 and head not in ignore:
                testcase.assertIn(head, st2)
                msg = ("AssertionError for header '%s' with format '%s': "
                       "%s and %s not equal within 2 places")
                testcase.assertAlmostEqual(
                    st1[head], st2[head], 2,
                    msg=msg % (head, format, st1[head], st2[head])
                )
        if len(ignore) == 0 or format != 'q':
            testcase.assertEqual(stream1[0].id, stream2[0].id)


class RFStreamTestCase(unittest.TestCase):

    def setUp(self):
        self.event = read_events()[0]
        self.station = AttribDict({'latitude': 41.818 - 66.7,
                                   'longitude': 79.689,
                                   'elevation': 365.4})

    def test_read_rf(self):
        self.assertIsInstance(read_rf(), RFStream)

    def test_io_header(self):
        stream = RFStream(read())[:1]
        for tr in stream:
            tr.stats.location = '11'
            tr.stats.pop('response', None)
        write_test_header(stream)
        test_io_header(self, stream)

    def test_io_header_obspy_stream(self):
        stream = read()[:1]
        for tr in stream:
            tr.stats.pop('response', None)
        ignore = ('back_azimuth', 'inclination')
        test_io_header(self, stream, ignore=ignore)

    def test_io_header_rf(self):
        stream = minimal_example_rf()[:1]
#        for tr in stream:
#            tr.stats.pop('sac')
        test_io_header(self, stream)

    def test_obj2stats(self):
        stats = obj2stats(event=self.event, station=self.station)
        for head, _ in _STATION_GETTER + _EVENT_GETTER:
            self.assertIn(head, stats)

    def test_rfstats(self):
        stats = rfstats(station=self.station, event=self.event, pp_depth=100.)
        for head in _HEADERS:
            if head not in _HEADERS_NOT_BY_RFSTATS:
                self.assertIn(head, stats)
        # event is exactly north from station and around 66.7 degrees away
        self.assertTrue(abs(stats.distance - 66.7) < 1.)
        self.assertTrue(abs(stats.back_azimuth % 360.) < 0.1)
        self.assertTrue(abs(stats.slowness - 6.4) < 0.1)
        # test issue 7
        event = self.event.copy()
        event.preferred_origin_id = None
        event.origins = []
        with self.assertRaisesRegex(ValueError, 'No origin'):
            stats = rfstats(station=self.station, event=event, pp_depth=100.)
        event = self.event.copy()
        event.preferred_magnitude_id = None
        event.magnitudes = []
        with self.assertRaisesRegex(ValueError, 'No magnitude'):
            stats = rfstats(station=self.station, event=event, pp_depth=100.)
        event = self.event.copy()
        del event.origins[0].latitude
        with self.assertRaisesRegex(ValueError, 'No origin'):
            stats = rfstats(station=self.station, event=event, pp_depth=100.)
        event = self.event.copy()
        del event.origins[0].depth
        with self.assertRaisesRegex(ValueError, 'No origin'):
            stats = rfstats(station=self.station, event=event, pp_depth=100.)

    def test_trim2(self):
        stream = read_rf()
        starttimes = [tr.stats.starttime for tr in stream]
        stream.trim2(50, 100, 'starttime')
        for t0, tr in zip(starttimes, stream):
            self.assertEqual(tr.stats.starttime - t0, 50)
            self.assertEqual(tr.stats.endtime - t0, 100)

    def test_slice2(self):
        stream = read_rf()
        endtimes = [tr.stats.endtime for tr in stream]
        stream = stream.slice2(-50, -20, 'endtime')
        for t0, tr in zip(endtimes, stream):
            self.assertEqual(tr.stats.starttime - t0, -50)
            self.assertEqual(tr.stats.endtime - t0, -20)

    def test_minimal_example_rf(self):
        stream = minimal_example_rf()
#        stream.select(component='L').plot_rf()
#        stream.select(component='Q').plot_rf()
#        stream.select(component='T').plot_rf()
#        import matplotlib.pyplot as plt
#        plt.show()
        stack = stream.stack()
        L = stack.select(component='L')
        Q = stack.select(component='Q')
        # check that maximum in L component is at 0s (at P onset)
        onset = L[0].stats.onset - L[0].stats.starttime
        dt = L[0].stats.delta
        self.assertAlmostEqual(L[0].data.argmax() * dt - onset, 0, delta=0.01)
        # check that maximum in Q component is at 8.6s
        # (subducting slab, Northern Chile)
        Q.trim2(5, 50, 'onset')
        onset = Q[0].stats.onset - Q[0].stats.starttime
        self.assertAlmostEqual(Q[0].data.argmax() * dt - onset, 8.6, delta=0.1)

    def test_minimal_example_Srf(self):
        stream = minimal_example_Srf()
#        stream.select(component='L').plot_rf()
#        stream.select(component='Q').plot_rf()
#        stream.select(component='T').plot_rf()
#        import matplotlib.pyplot as plt
#        plt.show()
        stack = stream.stack()
        L = stack.select(component='L')
        Q = stack.select(component='Q')
        # check that maximum in Q component is at 0s (at S onset)
        onset = Q[0].stats.onset - Q[0].stats.starttime
        dt = Q[0].stats.delta
        self.assertAlmostEqual(Q[0].data.argmax() * dt - onset, 0, delta=0.01)
        # check that maximum in L component is at 8.6s
        L.trim2(-5, 16, 'onset')
        onset = L[0].stats.onset - L[0].stats.starttime
        self.assertAlmostEqual(L[0].data.argmax() * dt - onset, 8.6, delta=0.1)

    def test_polarity_R_component(self):
        """issue #4"""
        stream = read_rf()
        rfstats(stream)
        stream.filter('bandpass', freqmin=0.5, freqmax=2)
        stream.trim2(10, 110, reftime='starttime')
        stream.rf(rotate='NE->RT')
        for tr in stream.select(component='R'):
            onset = tr.stats.onset - tr.stats.starttime
            dt = tr.stats.delta
            self.assertAlmostEqual(tr.data.argmax() * dt - onset, 0,
                                   delta=0.01)

    def test_str(self):
        s = ('Prf CX.PB01..BHT | -10.0s - 80.0s onset:'
             '2011-02-25T13:15:38.169539Z | 5.0 Hz, 451 samples | '
             'mag:6.0 dist:46.1 baz:325.0 slow:6.40 (Ps moveout)')
        stream = minimal_example_rf()
        self.assertEqual(str(stream[0]), s)

    def test_add_processing(self):
        stream = minimal_example_rf()
        proc = ' '.join(stream[0].stats.processing)
        self.assertIn('deconvolve(', proc)
        self.assertIn('rf(', proc)


def suite():
    return unittest.makeSuite(RFStreamTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



import unittest
import warnings
from rfsed import ReverbFilter
from rfsed.ReverbFilter import Resonance_Filt, plotfiltrf
from rfsed.hkYu import hkYu, plothkYu
import numpy as np
from rfsed.util import rfMoho_example, save_plot, save_tests
try:
    import matplotlib
except ImportError:
    Matplotlib = None

class ReverbFilter(unittest.TestCase):
    def test_Resonance_Filt(self):
        rfstream = rfMoho_example()
        self.assertTrue(Resonance_Filt(rfstream))

    def test_hkYu(self):
        rfstream = rfMoho_example()
        FilteredRF= Resonance_Filt(rfstream)
        hkYu(FltResult=FilteredRF, rayp=0.04, HSubSed=np.linspace(20,60,201), KSubSed=np.linspace(1.65,1.95,121), 
               HSed=np.linspace(0,10,201), KSed=np.linspace(1.65,2.25,201), VpMoho=6.9, VpSed= 2.5,  
               w1SubSed=0.6, w2SubSed=0.3, w3SubSed=0.1, w1Sed=0.6, w2Sed=0.3, w3Sed=0.1)  
    def test_plothkYu(self):
        rfstream = rfMoho_example()
        FilteredRF= Resonance_Filt(rfstream)
        savepath=save_tests()
        HKResults=hkYu(FltResult=FilteredRF, rayp=0.04, HSubSed=np.linspace(20,60,201), KSubSed=np.linspace(1.65,1.95,121), 
                HSed=np.linspace(0,10,201), KSed=np.linspace(1.65,2.25,201), VpMoho=6.9, VpSed= 2.5,  
                w1SubSed=0.6, w2SubSed=0.3, w3SubSed=0.1, w1Sed=0.6, w2Sed=0.3, w3Sed=0.1)  
        plothkYu(hkYuResult=HKResults, savepath=savepath)
        
def suite():
    return unittest.makeSuite(ReverbFilter, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')


#%% Test the Waveform Fitting Function
import os
import numpy as np
from rfsed.rf import read_rf
from rfsed.WaveformFittingMultiproc import WaveformPara, run_waveformfitting, plotbestmodel
from rfsed.util import rfMoho_example, rfSed_example, save_plot

#----------------------------------------------------------
savedir=save_plot()
savepath = savedir + '/WaveformFittingMultiproc/'
if not os.path.exists(savepath):  # create data folder if necessary
    os.mkdir(savepath)

rfstream= rfMoho_example()
VpSed=3.3
Sedthick=1.8
VpCrust = 6.9
gauparameter=1.25
KMoho= np.linspace(1.65,1.95,121)
HMoho=np.linspace(20,60,201)
# KMoho= np.linspace(1.65,1.95,5)
# HMoho=np.linspace(20,60,5)
rayp=0.04
delta = rfstream[0].stats.delta
wtCorr=0.5
wtRMSE=0.3
wtPG=0.2
format='pdf'
nproc=30
ModelParams=WaveformPara(rfstream, Sedthick, VpSed, VpCrust, rayp, KMoho, HMoho, gauparameter)
Results=run_waveformfitting(nproc, HMoho, ModelParams)
plotbestmodel(Results, ModelParams, wtCorr, wtRMSE, wtPG, savepath, format)


import unittest
from rfsed import ExtractEqMultiproc
from rfsed.ExtractEqMultiproc import Add_sta_time_2cat, Get_Eqtimes_multiproc, ExtractEq_Multiproc
from rfsed.util import raw_data_example, save_tests, catalog_example
from glob import glob
from obspy import read_events

class ExtractEqMultiproc(unittest.TestCase):
    def test_Add_sta_time_2cat(self):
        catalog = catalog_example()
        Request_window=[-50, 150]
        stalat=6.183432
        stalon=53.48334
        Add_sta_time_2cat(catalogfile=catalog, stalat=stalat, stalon=stalon, Request_window=Request_window)
    def test_Get_Eqtimes_multiproc(self):
        catalog = catalog_example()
        Request_window=[-50, 150]
        stalat=6.183432
        stalon=53.48334
        catalog_time = Add_sta_time_2cat(catalogfile=catalog, stalat=stalat, stalon=stalon, Request_window=Request_window)
        nproc=25
        Get_Eqtimes_multiproc(catalog=catalog_time, nproc=nproc)

    def test_ExtractEq_Multiproc(self):
        datapath=raw_data_example()
        datafiles=datafiles = sorted(glob("%s*.dat"%(datapath)))
        staname='NE301'
        catalog = catalog_example()
        Request_window=[-50, 150]
        stalat=6.183432
        stalon=53.48334
        nproc=25
        savepath=save_tests()
        filename = savepath + 'eq_data_multiproc.h5'
        catalog_time = Add_sta_time_2cat(catalogfile=catalog, stalat=stalat, stalon=stalon, Request_window=Request_window)
        timewindow=Get_Eqtimes_multiproc(catalog=catalog_time, nproc=nproc)
        ExtractEq_Multiproc(datafiles=datafiles, nproc=nproc, filename=filename,  timewindow=timewindow)
        
        
def suite():
    return unittest.makeSuite(ExtractEqMultiproc, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')


import unittest
import warnings
from rfsed import ReverbFilter
from rfsed.ReverbFilter import Resonance_Filt, plotfiltrf
import numpy as np
from rfsed.util import rfMoho_example, save_plot, save_tests
try:
    import matplotlib
except ImportError:
    Matplotlib = None

class ReverbFilter(unittest.TestCase):
    def test_Resonance_Filt(self):
        rfstream = rfMoho_example()
        self.assertTrue(Resonance_Filt(rfstream))
    def test_plotfiltrf(self):
        rfstream = rfMoho_example()
        savepath=save_tests()
        FilteredRF= Resonance_Filt(rfstream)
        plotfiltrf(FilteredRF, savepath, format = 'jpg')        

def suite():
    return unittest.makeSuite(ReverbFilter, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')


import numpy as np
import os 
import matplotlib.pyplot as plt
from rfsed.SynRF.FwdRF import SynRF
from rfsed.util import save_plot
depth = np.array([2, 35, 77.5])
vp = np.array([2.0, 6.5, 8.045])
vs = np.array([1.36, 3.75, 4.485])
rho=np.array([2.72, 2.92, 3.3455])
delay=5
n=2100
rayp=0.04
gaussian=1.25
delta=0.025
Synth=SynRF(depth, vp, vs, rho, rayp, dt=delta, npts=n, ipha=1)
Synth.run_fwd()
Synth.filter(freqmin=0.05, freqmax=1.25, order=2, zerophase=True)
rf_synth=Synth.run_deconvolution(pre_filt=[0.05, 1.25], shift=delay, gaussian=gaussian)
trdata=(rf_synth[0]).data
tlen = len(trdata)
t = np.arange(0, tlen)
t = (delta *  t) - delay # shift the time axis
fig = plt.figure(figsize=(11,8))
ax1 = fig.add_subplot(111)
ax1.plot(t, trdata, color = 'blue')
savedir=save_plot()
savepath = savedir + '/Example_SyntheticRF/'
if not os.path.exists(savepath):  # create data folder if necessary
    os.mkdir(savepath)
savepath=savepath + 'SyntheticRF.jpeg'
plt.savefig(savepath , format='jpeg', transparent=False,\
            dpi=250, bbox_inches = 'tight', pad_inches=0.1)


"""
.. module:: WaveformFitting
        :synopsis: Waveform Fitting for Moho depth and Vp/Vs ratio
.. moduleauthor:: Stephen Akinremi <s.akinremi@utwente.nl> Islam Fadel <i.e.a.m.fadel@utwente.nl> (October 2023)
"""

import numpy as np
from rfsed.SynRF.FwdRF import SynRF
import math
from obspy.signal.tf_misfit import pg
import matplotlib.pyplot as plt
from matplotlib.legend_handler import HandlerLine2D as HL 
from sklearn.metrics import mean_squared_error
import matplotlib.gridspec as gridspec
import matplotlib.transforms as mtransforms

def WaveformFitting(rfstream, HSed, VpSed, VpCrust, rayp, KMoho, HMoho, gaussian, wtCorr, wtRMSE, wtPG, savepath, format):
    """
    This function performs waveform fitting for the Moho depth and Vp/Vs ratio
    Returns the best model based on the Correlation Coefficient(CC), Root-mean-squared error(RMSE), 
    Phase-goodness-of-fit(PG) and Overall Goodness-of-fit that combines CC, RMSE, and PG
    Weights of CC, RMSE and PG (wtCorr,wtRMSE, wtPG) are given as input and can be modified before plotting
    At all times, wtCorr+wtRMSE+wtPG=1.0

    :param rfstream: Obspy RFStream object containing the receiver functions
    :type rfstream: Obspy RFStream object
    :param HSed: Sediment thickness in km
    :type HSed: float
    :param VpSed: Sediment Vp in km/s
    :param VpSed: float
    :param VpCrust: Crustal Vp in km/s
    :param VpCrust: float
    :param rayp: Ray parameter in s/km
    :param rayp: float
    :param KMoho: Vp/Vs ratio range for Moho
    :type KMoho: numpy array
    :param HMoho: Moho depth range in km
    :type HMoho: numpy array
    :param gaussian: Gaussian filter parameter (a)
    :type gaussian: float
    :param wtCorr: Weight of Correlation Coefficient
    :type wtCorr: float
    :param wtRMSE: Weight of Root-mean-squared error
    :type wtRMSE: float
    :param wtPG: Weight of Phase Goodness of Fit
    :type wtPG: float
    :param savepath: Path to save the plots
    :type savepath: str
    :param format: Format of the plots
    :type format: str
    :return: Dictionary containing the Waveform Fitting Results
    """
    VsSed=0.7858 - 1.2344*VpSed + 0.7949*VpSed**2 - 0.1238*VpSed**3 + 0.0064*VpSed**4
    SedDen=((1.6612*(VpSed)) - (0.4721 * ((VpSed)**2)) + (0.0671* ((VpSed)**3)) - (0.0043* ((VpSed)**4)) + (0.000106* ((VpSed)**5)))
    VsCrust=0.7858 - 1.2344*VpCrust + 0.7949*VpCrust**2 - 0.1238*VpCrust**3 + 0.0064*VpCrust**4
    CrustDen=((1.6612*(VpCrust)) - (0.4721 * ((VpCrust)**2)) + (0.0671* ((VpCrust)**3)) - (0.0043* ((VpCrust)**4)) + (0.000106* ((VpCrust)**5)))
    #----------------------------------------------------------
    # RF Stream 
    delta = rfstream[0].stats.delta
    # print('Data delta is', delta)
    b = rfstream[0].stats.starttime - rfstream[0].stats.onset
    t = (np.arange(0, len(rfstream[0].data), 1) * delta) + b
    staname = rfstream[0].stats.station
    Stacked=rfstream.select(component='R').stack()
    StackedRF=Stacked[0]
    StackedData=StackedRF.data
    Stdd=np.std(StackedData, axis=0)
    StddPlus = StackedData + Stdd
    StddMinus = StackedData - Stdd
    delay = abs(t[0])
    n=len(StackedData)
    # Waveform Fitting start from t = 0 to t = 25 seconds
    tzero=np.where(t==0)[0]
    tzero=int(tzero)  
    start=tzero
    end= 25/delta
    end=tzero + int(end)
    #----------------------------------------------------------
    MohoSol=[]
    z=0
    #----------------------------------------------------------
    for i in range(len(KMoho)):
        Ktemp = KMoho[i]
        for j in range(len(HMoho)): 
            Htemp = HMoho[j]
            #----------------------------------------------------------
            # Calculate the model parameters
            #----------------------------------------------------------
            depth = np.array([HSed, Htemp, 77.5])
            vp = np.array([VpSed, VpCrust, 8.045])
            vs = np.array([VsSed, (VpCrust/Ktemp), 4.49])
            rho = np.array([SedDen, CrustDen, 3.299])
            Synth=SynRF(depth, vp, vs, rho, rayp, dt=delta, npts=n, ipha=1)
            Synth.run_fwd()
            Synth.filter(freqmin=0.05, freqmax=1.25, order=2, zerophase=True)
            rf_synth=Synth.run_deconvolution(pre_filt=[0.05, 1.25], shift=delay, gaussian=gaussian)
            rf_synth=(rf_synth[0]).data
            StackedData = StackedData
            #----------------------------------------------------------
            # Goodness of Fit Evaluation
            #----------------------------------------------------------
            # general constants for goodness of fit evalaution
            dt = delta
            fmin = .5
            fmax = 10
            nf = 100
            n=n
            #Subset the Real and Synthetic Data to 0 to 25 seconds
            RealData = StackedData[start:end]
            SynthData = rf_synth[start:end]
            t=t[start:end]
            #----------------------------------------------------------
            st1a= RealData
            st2= SynthData  # st2 is the reference data
            #----------------------------------------------------------
            # Correlation Coefficient
            Correlation = np.corrcoef(RealData, SynthData)
            Correlation = Correlation[0,1]
            #----------------------------------------------------------
            # RMSE Criteria
            RMSE = math.sqrt(mean_squared_error(RealData, SynthData))
            #----------------------------------------------------------
            # Normalise Phase Goodness of Fit Criteria (PG)
            abs_PG=abs(RealData)
            abs_rf=abs(SynthData)
            PG_Eq=np.where(abs_PG == 0, abs_PG, RealData/abs_PG)
            Synth_Eq=np.where(abs_rf == 0, abs_rf, SynthData/abs_rf)
            PG=pg(PG_Eq, Synth_Eq, dt=dt, fmin=fmin, fmax=fmax, nf=nf, w0=6, st2_isref=True, a=10., k=1.)
            #----------------------------------------------------------
            MohoSol.append([Correlation, RMSE, PG, Htemp, Ktemp])
    AllSolutions = np.zeros((len(MohoSol), 5))
    for i in MohoSol:
        AllSolutions[z,:] = [i[0], i[1], i[2], i[3], i[4]]
        z=z+1
    #----------------------------------------------------------
    # Normalise the Correlation and RMSE and PG values 
    #----------------------------------------------------------
    Corr=AllSolutions[:,0]
    MaxCorr=max(Corr)
    RMSE=AllSolutions[:,1]
    MaxRSME=max(RMSE)
    PG=AllSolutions[:,2]
    MaxPG=max(PG)
    NormCorr=Corr/MaxCorr
    NormRMSE=RMSE/MaxRSME
    NormPG=PG/MaxPG
    GoF= wtCorr*NormCorr + wtRMSE*(1-NormRMSE) + wtPG*NormPG #Invert the RMSE values
    #-------------------------
    Htemp=AllSolutions [:,3]
    Ktemp=AllSolutions [:,4]
    # #-------------------------
    AllCriteria=zip(Corr, RMSE, PG, GoF, Htemp, Ktemp)
    WaveformSolutions = np.zeros((len(GoF), 6))
    z=0
    for i in AllCriteria:
        WaveformSolutions[z,:] = [i[0], i[1], i[2], i[3], i[4], i[5]]
        z=z+1
    BestModel_Correlation = WaveformSolutions[np.argmax(WaveformSolutions[:,0]),:]
    BestModel_RMSE = WaveformSolutions[np.argmin(WaveformSolutions[:,1]),:]
    BestModel_PG = WaveformSolutions[np.argmax(WaveformSolutions[:,2]),:]
    BestModel_GoF = WaveformSolutions[np.argmax(WaveformSolutions[:,3]),:]
    print('Best Model from Correlation Coefficient: Moho at %.2f km with Vp/Vs of %.2f'%(BestModel_Correlation[4],BestModel_Correlation[5]))
    print('Best Model from RMSE: Moho at %.2f km with Vp/Vs of %.2f'%(BestModel_RMSE[4],BestModel_RMSE[5]))
    print('Best Model from PG: Moho at %.2f km with Vp/Vs of %.2f'%(BestModel_PG[4],BestModel_PG[5]))
    print('Best Model from Overall Goodness of Fit: Moho at %.2f km with Vp/Vs of %.2f'%(BestModel_GoF[4],BestModel_GoF[5]))
    WaveformFitting={'WaveformSolutions':WaveformSolutions, 'rfstream':rfstream, 'HSed':HSed, 'VpSed':VpSed, 'VpCrust':VpCrust, 
                     'rayp':rayp, 'KMoho':KMoho, 'HMoho':HMoho, 'gaussian':gaussian, 'wtCorr':wtCorr, 'wtRMSE':wtRMSE, 'wtPG':wtPG, 
                     'savepath':savepath, 'format':format}
    return WaveformFitting

def PlotWaveformFitting(WaveformFitting, wtCorr, wtRMSE, wtPG, savepath, format):
    """
    This function plots the waveform fitting results for the Moho depth and Vp/Vs ratio
    Plots the Waveform Fitting Results for CC, RMSE, PG and GoF
    Weights of CC, RMSE and PG are given as input and can be modified before plotting
    At all times, wtCorr+wtRMSE+wtPG=1.0

    :param WaveformFitting: Dictionary containing the Waveform Fitting Results from the function WaveformFitting
    :type WaveformFitting: dict
    :param wtCorr: Weight of Correlation Coefficient
    :type wtCorr: float
    :param wtRMSE: Weight of Root-mean-squared error
    :type wtRMSE: float
    :param wtPG: Weight of Phase Goodness of Fit
    :type wtPG: float
    :param savepath: Path to save the plots
    :type savepath: str
    :param format: Format of the plots
    :type format: str
    """
    WaveformSolutions=WaveformFitting['WaveformSolutions']
    rfstream=WaveformFitting['rfstream']
    staname = rfstream[0].stats.station
    HSed=WaveformFitting['HSed']
    VpSed=WaveformFitting['VpSed']
    VpCrust=WaveformFitting['VpCrust']
    rayp=WaveformFitting['rayp']
    KMoho=WaveformFitting['KMoho']
    HMoho=WaveformFitting['HMoho']
    gaussian=WaveformFitting['gaussian']
    BestModel_Correlation = WaveformSolutions[np.argmax(WaveformSolutions[:,0]),:]
    BestModel_RMSE = WaveformSolutions[np.argmin(WaveformSolutions[:,1]),:]
    BestModel_PG = WaveformSolutions[np.argmax(WaveformSolutions[:,2]),:]
    BestModel_GoF = WaveformSolutions[np.argmax(WaveformSolutions[:,3]),:]
    Corr = WaveformSolutions[:,0]
    RMSE = WaveformSolutions[:,1]
    PG = WaveformSolutions[:,2]
    GoF = WaveformSolutions[:,3]
    #----------------------------------------------------------
    labels=['Correlation Coefficient', 'Root-mean-square Error', 'Phase Goodness of Fit', 'Overall Goodness of Fit']
    model_criteria=zip(Corr, RMSE, PG, GoF)
    print(model_criteria)
    for i in model_criteria:
        print(i)
    for i in range(len(labels)):
        suff=labels[i]
        if i==0:
            Criteria=Corr
            BestModel=BestModel_Correlation
        elif i==1:
            Criteria=RMSE
            BestModel=BestModel_RMSE
        elif i==2:
            Criteria=PG
            BestModel=BestModel_PG
        elif i==3:
            Criteria=GoF
            BestModel=BestModel_GoF
        minMoho=HMoho[0]
        maxMoho=HMoho[-1]
        minVpVs=KMoho[0]
        maxVpVs=KMoho[-1]
        # ----------------------------------------------------------
        # Plot the Solutions
        # ----------------------------------------------------------
        plt.figure(figsize=(18, 6)) 
        gs_corr = gridspec.GridSpec(1, 3, width_ratios=[6,8,4])
        ax1 = plt.subplot(gs_corr[0, 0])
        plt.tricontourf(WaveformSolutions[:,4], WaveformSolutions[:,5], Criteria,50, cmap ="jet")
        cb=plt.colorbar(format='%.3f', orientation="vertical")
        cb.ax.set_xlabel('S', rotation=0)
        p1,=ax1.plot(BestModel[4],BestModel[5], 'k+', color='black', mew=2, ms=10, \
        label='%s Best Model %.2f km %.2f Vp/Vs'%(staname,BestModel[4],BestModel[5]+0.001))
        props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
        plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,\
            ncol=2, mode="expand", borderaxespad=0, \
            handler_map={p1:HL(numpoints=1)})
        ax1.set_ylabel('Vp/Vs')
        ax1.set_xlabel('depth (km)')
        ax1.set_ylim(minVpVs, maxVpVs)
        ax1.set_xlim(minMoho, maxMoho)
        #----------------------------------------------------------
        # Plot Stacked Data, Standard Deviation and Synthetic RF )(Show time between -5 to 25 seconds)
        #----------------------------------------------------------
        delta = rfstream[0].stats.delta
        b = rfstream[0].stats.starttime - rfstream[0].stats.onset
        t = (np.arange(0, len(rfstream[0].data), 1) * delta) + b
        delay=abs(t[0]) 
        staname = rfstream[0].stats.station
        Stacked=rfstream.select(component='R').stack()
        StackedRF=Stacked[0]
        StackedData=StackedRF.data
        Stdd=np.std(StackedData, axis=0)
        StddPlus = StackedData + Stdd
        StddMinus = StackedData - Stdd
        n=len(StackedData)
        #----------------------------------------------------------
        H=BestModel[4]
        K=BestModel[5]
        # Calculate sediment and crustal density using Nafe-Drake curve (Brocher Theorem) valid for vp between 1.5 and 8.5 km/sec
        VsSed=0.7858 - 1.2344*VpSed + 0.7949*VpSed**2 - 0.1238*VpSed**3 + 0.0064*VpSed**4
        SedDen = ((1.6612*(VpSed)) - (0.4721 * ((VpSed)**2)) + (0.0671* ((VpSed)**3)) - (0.0043* ((VpSed)**4)) + (0.000106* ((VpSed)**5)))
        CrustDen = ((1.6612*(VpCrust)) - (0.4721 * ((VpCrust)**2)) + (0.0671* ((VpCrust)**3)) - (0.0043* ((VpCrust)**4)) + (0.000106* ((VpCrust)**5)))
        depth = np.array([HSed, H, 77.5])
        vp = np.array([VpSed, VpCrust, 8.045])
        vs = np.array([VsSed, (VpCrust/K), 4.49])
        rho = np.array([SedDen, CrustDen, 3.299])
        Synth=SynRF(depth, vp, vs, rho, rayp, dt=delta, npts=n, ipha=1)
        Synth.run_fwd()
        Synth.filter(freqmin=0.05, freqmax=1.25, order=2, zerophase=True)
        rf_synth=Synth.run_deconvolution(pre_filt=[0.05, 1.25], shift=delay, gaussian=gaussian)
        rf_synth=(rf_synth[0]).data
        l = len(rf_synth)
        synth_time = np.arange(0, l)
        synth_time = (delta *  synth_time) - delay 
        # # Plot to visualise difference
        ax2=plt.subplot(gs_corr[0, 1])
        plt.plot(t, StackedData, color = 'black', label = 'Stacked RF')
        plt.plot(synth_time, rf_synth, '--',  color = 'black', label = 'Synthetic RF')
        plt.fill_between(t, StddPlus, StddMinus, alpha = 0.3, color = 'grey', 
                         label = 'RF Data Std' )
        # p1,=ax2.plot(BestModel[4],BestModel[5], 'k+', color='black', mew=2, ms=10, \
        # label='%s Best Model'%(suff))
        ax2.set_title("%s Best Model" %suff)
        # Subset the Data to -5 to 25 seconds
        ax2.set_xlabel('time (s)')
        ax2.set_ylabel('amplitude')
        ax2.set_yticklabels([])
        ax2.set_yticks([])
        ax2.set_xlim(-5, 25)
        ax2.legend()
        # #----------------------------------------------------------
        # # Plot Vp and Vs with Depth
        # #----------------------------------------------------------
        ax3=plt.subplot(gs_corr[0, 2])
        dep=(0, HSed, HSed, H, H, 50)
        vp=(VpSed, VpSed, VpCrust, VpCrust, 8.045, 8.045)
        vs=(VsSed, VsSed, VpCrust/K, VpCrust/K, 4.49, 4.49)
        ax3.plot(vp, dep, '--', color = 'red', label = 'Vp')
        ax3.plot(vs, dep, ':', color = 'blue', label = 'Vs')
        ax3.xaxis.set_ticks_position('bottom')
        ax3.set_ylim(0, 50)                   
        ax3.set_xlim(1, 10)  
        ax3.invert_yaxis()
        ax3.set_ylabel('depth(km)')
        ax3.legend(fontsize=15)
        ax3.set_xlabel("v (km/s)")#, fontsize=20)
        #------------------------------------------
        #    Save Plot to File
        #-------------------------------------------
        filename = "%s/%s_%s.%s"%(savepath, staname, suff, format)
        plt.savefig(filename , format=format, transparent=False, 
                    dpi=300, bbox_inches = 'tight', pad_inches=0.1)
        plt.show()
        plt.close("all")  
#----------------------------------------------------------


#%% Test the Resonance filtering and Modified HK Stacking of Yu et al 2015
import os 
import numpy as np
from obspy import read
from rfsed.ReverbFilter import Resonance_Filt, plotfiltrf
from rfsed.util import rfMoho_example, rfSed_example, save_plot

savedir=save_plot()
savepath = savedir + '/Resonance_Filtering/'
if not os.path.exists(savepath):  # create data folder if necessary
    os.mkdir(savepath)
rfstream= rfMoho_example()
VpSed=2.1
VpMoho=6.9
SedH= 0.6
VsSed= 0.7858 - 1.2344*VpSed + 0.7949*VpSed**2 - 0.1238*VpSed**3 + 0.0064*VpSed**4
gaussalp=1.25
rayp = 0.04
w1Sed, w2Sed, w3Sed = [0.6, 0.3, 0.1]
w1SubSed, w2SubSed, w3SubSed = [0.6, 0.3, 0.1]
#----------------------------------------------------------
FilteredRF= Resonance_Filt(rfstream)
plotfiltrf(FilteredRF, savepath, format = 'jpg')

#%% Test the Modified HK Stacking of Yu et al 2015
# Requires result from Resonance filter 
import os 
import numpy as np
from obspy import read
from rfsed.hkYu import hkYu, plothkYu
from rfsed.util import rfMoho_example, rfSed_example, save_plot

savedir=save_plot()
savepath = savedir + '/HK_Yu_Method/'
if not os.path.exists(savepath):  # create data folder if necessary
    os.mkdir(savepath)
rfstream= rfMoho_example()
VpSed=2.1
VpMoho=6.9
SedH= 0.6
VsSed= 0.7858 - 1.2344*VpSed + 0.7949*VpSed**2 - 0.1238*VpSed**3 + 0.0064*VpSed**4
gaussalp=1.25
rayp = 0.04
w1Sed, w2Sed, w3Sed = [0.6, 0.3, 0.1]
w1SubSed, w2SubSed, w3SubSed = [0.6, 0.3, 0.1]
#----------------------------------------------------------
FilteredRF= Resonance_Filt(rfstream)
# plotfiltrf(FilteredRF, savepath, format = 'jpg')
HKResults=hkYu(FltResult=FilteredRF, rayp=0.04, HSubSed=np.linspace(20,60,201), KSubSed=np.linspace(1.65,1.95,121), 
               HSed=np.linspace(0,10,201), KSed=np.linspace(1.65,2.25,201), VpMoho=6.9, VpSed= 2.5,  
               w1SubSed=0.6, w2SubSed=0.3, w3SubSed=0.1, w1Sed=0.6, w2Sed=0.3, w3Sed=0.1)
plothkYu(hkYuResult=HKResults, savepath=savepath, g = [75.,10., 15., 2.5], rmneg = None, format = 'jpg')


# Copyright 2013-2016 Tom Eulenfeld, MIT license
"""
Tests for profile module.
"""
import unittest

import numpy as np
from rfsed.rf.profile import get_profile_boxes
from rfsed.rf.tests.test_rfstream import test_io_header
from rfsed.rf.util import minimal_example_rf

try:
    import cartopy
except ImportError:
    cartopy = None


class ProfileTestCase(unittest.TestCase):

    @unittest.skipIf(cartopy is None, 'cartopy not installed')
    def test_profile(self):
        dx = np.linspace(-50, 50, 41)
        boxes = get_profile_boxes((-21, -69.5), 85, dx, width=300)
        stream = minimal_example_rf()
        stream.extend(stream[:3])  # to actually stack something
        profile = stream.select(component='Q').profile(boxes)
        self.assertEqual(len(profile), 3)
        str_ = ('Prf profile (Q) | -10.0s - 80.0s | 5.0 Hz, 451 samples | '
                'pos:-6.25km slow:6.40 (Ps moveout)')
        self.assertEqual(str(profile[0]), str_)
        test_io_header(self, profile[:1])
        self.assertIn('profile(', ' '.join(profile[0].stats.processing))
        # test plots
        profile.plot_profile(top='hist')
        from rfsed.rf.imaging import plot_profile_map
        plot_profile_map(boxes)


def suite():
    return unittest.makeSuite(ProfileTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# Copyright 2013-2019 Tom Eulenfeld, MIT license
"""
All objects and functions below are after the Tom Eulenfeld's
rf project License below.

The MIT License (MIT)

Copyright (c) 2013-2019 Tom Eulenfeld

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
"""
Utility functions and classes for receiver function calculation.
"""
import collections
import inspect
import itertools
from pkg_resources import resource_filename

from decorator import decorator
import numpy as np


DEG2KM = 111.2  #: Conversion factor from degrees epicentral distance to km


def _get_stations(inventory):
    channels = inventory.get_contents()['channels']
    stations = {ch[:-1] + '?': ch[-1] for ch in channels}
    return stations


def iter_event_data(events, inventory, get_waveforms, phase='P',
                    request_window=None, pad=10, pbar=None, **kwargs):
    """
    Return iterator yielding three component streams per station and event.

    :param events: list of events or `~obspy.core.event.Catalog` instance
    :param inventory: `~obspy.core.inventory.inventory.Inventory` instance
        with station and channel information
    :param get_waveforms: Function returning the data. It has to take the
        arguments network, station, location, channel, starttime, endtime.
    :param phase: Considered phase, e.g. 'P', 'S', 'PP'
    :type request_window: tuple (start, end)
    :param request_window: requested time window around the onset of the phase
    :param float pad: add specified time in seconds to request window and
       trim afterwards again
    :param pbar: tqdm_ instance for displaying a progressbar
    :param kwargs: all other kwargs are passed to `~rf.rfstream.rfstats()`

    :return: three component streams with raw data

    Example usage with progressbar::

        from tqdm import tqdm
        from rf.util import iter_event_data
        with tqdm() as t:
            for stream3c in iter_event_data(*args, pbar=t):
                do_something(stream3c)

    .. _tqdm: https://pypi.python.org/pypi/tqdm
    """
    from rfsed.rf.rfstream import rfstats, RFStream
    method = phase[-1].upper()
    if request_window is None:
        request_window = (-50, 150) if method == 'P' else (-100, 50)
    stations = _get_stations(inventory)
    if pbar is not None:
        pbar.total = len(events) * len(stations)
    for event, seedid in itertools.product(events, stations):
        if pbar is not None:
            pbar.update(1)
        origin_time = (event.preferred_origin() or event.origins[0])['time']
        try:
            args = (seedid[:-1] + stations[seedid], origin_time)
            coords = inventory.get_coordinates(*args)
        except Exception:  # station not available at that time
            continue
        try:
            stats = rfstats(station=coords, event=event, phase=phase, **kwargs)
        except Exception as ex:
            from warnings import warn
            warn('Error "%s" in rfstats call for event %s, station %s.'
                 % (ex, event.resource_id, seedid))
            continue
        if not stats:
            continue
        net, sta, loc, cha = seedid.split('.')
        starttime = stats.onset + request_window[0]
        endtime = stats.onset + request_window[1]
        kws = {'network': net, 'station': sta, 'location': loc,
               'channel': cha, 'starttime': starttime - pad,
               'endtime': endtime + pad}
        try:
            stream = get_waveforms(**kws)
        except Exception:  # no data available
            continue
        stream.trim(starttime, endtime)
        stream.merge()
        if len(stream) != 3:
            from warnings import warn
            warn('Need 3 component seismograms. %d components '
                 'detected for event %s, station %s.'
                 % (len(stream), event.resource_id, seedid))
            continue
        if any(isinstance(tr.data, np.ma.masked_array) for tr in stream):
            from warnings import warn
            warn('Gaps or overlaps detected for event %s, station %s.'
                 % (event.resource_id, seedid))
            continue
        for tr in stream:
            tr.stats.update(stats)
        yield RFStream(stream)


def iter_event_metadata(events, inventory, pbar=None):
    """
    Return iterator yielding metadata per station and event.

    :param events: list of events or `~obspy.core.event.Catalog` instance
    :param inventory: `~obspy.core.inventory.inventory.Inventory` instance
        with station and channel information
    :param pbar: tqdm_ instance for displaying a progressbar
    """
    stations = _get_stations(inventory)
    if events is None:
        events = [None]
    if pbar is not None:
        pbar.total = len(events) * len(stations)
    for event, seedid in itertools.product(events, stations):
        if pbar is not None:
            pbar.update(1)
        net, sta, loc, cha = seedid.split('.')
        meta = {'network': net, 'station': sta, 'location': loc,
                'channel': cha}
        if event is not None:
            ot = (event.preferred_origin() or event.origins[0])['time']
            meta['event_time'] = ot
        yield meta


class IterMultipleComponents(object):

    """
    Return iterable to iterate over associated components of a stream.

    :param stream: Stream with different, possibly many traces. It is
        split into substreams with the same seed id (only last character
        i.e. component may vary)
    :type key: str or None
    :param key: Additionally, the stream is grouped by the values of
         the given stats entry to differentiate between e.g. different events
         (for example key='starttime', key='onset')
    :type number_components: int, tuple of ints or None
    :param number_components: Only iterate through substreams with
         matching number of components.
    """

    def __init__(self, stream, key=None, number_components=None):
        substreams = collections.defaultdict(stream.__class__)
        for tr in stream:
            k = (tr.id[:-1], str(tr.stats[key]) if key is not None else None)
            substreams[k].append(tr)
        n = number_components
        self.substreams = [s for _, s in sorted(substreams.items())
                           if n is None or len(s) == n or
                           (not isinstance(n, int) and len(s) in n)]

    def __len__(self):
        return len(self.substreams)

    def __iter__(self):
        for s in self.substreams:
            yield s


def direct_geodetic(latlon, azi, dist):
    """
    Solve direct geodetic problem with geographiclib.

    :param tuple latlon: coordinates of first point
    :param azi: azimuth of direction
    :param dist: distance in km

    :return: coordinates (lat, lon) of second point on a WGS84 globe
    """
    from geographiclib.geodesic import Geodesic
    coords = Geodesic.WGS84.Direct(latlon[0], latlon[1], azi, dist * 1000)
    return coords['lat2'], coords['lon2']


__CACHE = {}


def minimal_example_rf():
    """
    Return receiver functions calculated from the data returned by read_rf().
    """
    cache_key = 'minimal_example_rf'
    if cache_key in __CACHE:
        return __CACHE[cache_key].copy()
    from rfsed.rf.rfstream import read_rf, rfstats
    stream = read_rf()
    rfstats(stream)
    stream.filter('bandpass', freqmin=0.5, freqmax=2)
    stream.trim2(10, 110, reftime='starttime')
    stream.rf(winsrc=(-5, 25, 5))
    stream.moveout()
    stream.trim2(-10, 80, reftime='onset')
    stream.ppoints(50)
    __CACHE[cache_key] = stream
    return stream.copy()


def minimal_example_Srf():
    """
    Return S receiver functions calculated from example data.
    """
    cache_key = 'minimal_example_Srf'
    if cache_key in __CACHE:
        return __CACHE[cache_key].copy()
    from rfsed.rf.rfstream import read_rf, rfstats
    fname = resource_filename('rfsed', 'rf/example/minimal_example_S.tar.gz')
    stream = read_rf(fname)
    rfstats(stream, phase='S')
    stream.filter('bandpass', freqmin=0.2, freqmax=0.5)
    stream.trim2(10, 120, reftime='starttime')
    stream.rf(method='S', winsrc=(-5, 15, 5))
    stream.moveout(phase='Sp')
    stream.ppoints(50, pp_phase='P')
    __CACHE[cache_key] = stream
    return stream.copy()


@decorator
def _add_processing_info(_func_, *args, **kwargs):
    from rfsed.rf import __version__
    args_ = inspect.getcallargs(_func_, *args, **kwargs)
    if args_.pop('self', None) is None:
        args_.pop('stream')
    kw = args_.pop('kwargs', {})
    kw.update(args_)
    boxes = kw.pop('boxes', None)
    if boxes:
        kw['latlon'] = boxes[0]['profile']['latlon']
        kw['azimuth'] = boxes[0]['profile']['azimuth']
        kw['length'] = boxes[0]['profile']['length']
    info = 'rf {version}: {function}(%s)'.format(
        version=__version__, function=_func_.__name__)
    arguments = ['%s=%s' % (k, repr(v)) if not isinstance(v, str) else
                 "%s='%s'" % (k, v) for k, v in kw.items()]
    info = info % '::'.join(sorted(arguments))
    stream = _func_(*args, **kwargs)
    try:
        for tr in stream:
            tr._internal_add_processing_info(info)
    except Exception:
        pass
    return stream



# Copyright (c) 2023 Stephen Akinremi MIT license
"""
rfsed Documentation
================

rfsed is a Python framework for analysizing receiver functions.
It is developed specifically to implement different techniques of analysing receiver functions
from stations overlying sedimentary layers. Receiver functions techniques implemented in rfsed are:
1. H-k stacking of Zhu and Kanamori (2000)
2. Sequential H-k stacking of Yeck et al., (2013)
3. Resonance filtering and modified H-k stacking of Yu et al., (2015)
4. H-k stacking and Waveform Fitting two-step method of Akinremi et al., (2024)
6. Analysis of the synthetic reciever functions with the above-mentioned methods.

Beside these receiver function methods, rfsed has the following features:
1. Creating publication quality figures for the results of the analysis.
2. Extracting earthquake data from local seismic record files.
3. Multiprocessing options for waveform fitting and extracting earthquake data from local seismic record files.


Receiver functions streams are handled by the 'RFStream' class of the rf open software, and it inherits a lot of useful methods from the Obspy class 'Stream'.
It is supported via the obspyh5 package. For more information on class "RFStream", see documentation on rf (https://rf.readthedocs.io/en/latest/). 
The rf framework is included in the rfsed package for completness under The MIT License (MIT) Copyright (c) 2013-2019 Tom Eulenfeld.
In the module to extract earthquake data from local seismic records, read and write support for necessary metadata is provided 
for SAC, SeismicHanlder and HDF5 formats based on ObsPy.

The forward receiver function calculation is done using SynR, which is modified after the SynSeis module of Mijian Xu's seispy project (under the GNU GENERAL PUBLIC LICENSE).

Method
------
The receiver function technique is a well-established technique to image velocity contrasts in the subsurface by using isolated P – to-S – wave 
conversions (P-receiver functions) and the reverberations generated at such discontinuities. The velocity contrast at the crust-mantle 
boundary (Moho) represents a major discontinuity that receiver functions are highly sensitive to, making them useful to constrain average 
crustal thickness and Vp/Vs. Several methods such as H-k stacking, has been developed to investigate the crustal thickness and Vp/Vs ratio 
using receiver functions. However, the presence of near-surface low-velocity sedimentary layer can obscure Moho signals due to the additional 
P-to-S wave conversions and  associated reverberated phases created at the sediment-basement discontinuity. These additional intra-crustal phases 
can have large amplitudes and similar arrival times to Moho signals, which makes it difficult to retrieve Moho information using standard RF techniques.
Several other methods have been developed to deal with the complex RF in sedimentary basins and recover crustal strucures (crustal thickness and Vp/Vs).
These methods include: Iterative trial and error fitting of observed receiver functions with synthetic ones,  wavefield continuation and decomposition technique, application of band-pass or resonance filters to remove the sediment signals in the receiver functions, 
sequential H-k stacking, and combination of H-k stacking and waveform fitting. 


Installation
------------

Dependencies of rfsed are
    * ObsPy_ and some of its dependencies,
    * cartopy, geographiclib, shapely,
    * mtspec_ for multitaper deconvolution,
    * toeplitz_ for faster time domain deconvolution (optional),
    * obspyh5_ for hdf5 file support,
    * tqdm for progress bar in batch processing (optional).

The easiest way to install rfsed is via `pip`_::
    
        pip install rfsed

Here are some instructions to install rfsed into a fresh conda environment::

    conda create -n rfsed python==3.9 
    conda activate rfsed
    git clone https://github.com/akinremisa/rfsed.git
    cd rfsed #Change directory to the same directory that this repo is in
    pip install -e .

Alternatively, the environment can be created and the package installed using the environment.yml file in the root directory of this repo::

1. Clone the rffw repository
git clone https://github.com/akinremisa/rfsed.git
2. Change directory to the same directory that this repo is cloned (i.e., same directory as environment.yml)
cd rfsed
3. Create the evniroment and install the rfsed package using the environment.yml file
conda env create -f environment.yml
4. Activate the environement
conda activate rfsed


Using the Python module/Tutorial
--------------------------------
Examples of how to use the methods in this Python module can be found in the `examples` directory.

Miscellaneous
-------------

Please feel free to request features, report bugs or contribute code on
GitHub_. The code is continuously tested by Github Actions.

Citation
--------

If you found this package useful, please consider citing it.
.....................................


.. _ObsPy: http://www.obspy.org/
.. _pip: http://www.pip-installer.org/
.. _obspyh5: https://github.com/trichter/obspyh5/
.. _toeplitz: https://github.com/trichter/toeplitz/
.. _mtspec: https://github.com/krischer/mtspec


.. _GitHub: https://github.com/akinremisa/rfsed/
"""



# %% Run the function to extract the data from local files
import os
from obspy import read
from os.path import exists
from os import mkdir
from rfsed.ExtractEq import ExtractEq
from rfsed.util import catalog_example, raw_data_example, save_Eq_data
staname='NE301'
stalat=6.183432
stalon=53.48334
datapath= raw_data_example()
savedir=save_Eq_data()
savepath=savedir+'/Extracted_Eq_Data/'
if not exists(savepath):
    mkdir(savepath)
filename = savepath + 'eq_data.h5'
catalog = catalog_example()
Request_window=[-50, 150] #Time relative to first P arrival
ExtractData= ExtractEq(datapath, filename, catalog, stalat, stalon, Request_window)
#%% Check the extracted data
ExtractedEQ_Data=read(filename)
print(len(ExtractedEQ_Data), ' traces in the extracted data')
print(ExtractedEQ_Data)
# %%



import unittest
import warnings
from rfsed.WaveformFitting import WaveformFitting, PlotWaveformFitting 
import numpy as np
from rfsed.util import rfMoho_example, save_tests
try:
    import matplotlib
except ImportError:
    Matplotlib = None

class TestWaveformFitting(unittest.TestCase):
    def test_WaveformFitting(self):
        savepath=save_tests()
        rfstream= rfMoho_example()
        VpSed=2.2
        HSed=0.1
        VpCrust = 6.9
        gaussian=1.25
        rayp=0.04
        # KMoho= np.linspace(1.65,1.95,121)
        # HMoho=np.linspace(20,60,201)
        KMoho= np.linspace(1.65,1.95,5)
        HMoho=np.linspace(20,60,5)
        wtCorr, wtRMSE, wtPG = [0.5, 0.3, 0.2] #wtCorr+wtRMSE+wtPG=1.0
        WaveformFitting(rfstream, HSed, VpSed, VpCrust, rayp, KMoho, HMoho, 
                     gaussian, wtCorr, wtRMSE, wtPG, savepath, format)

    def test_PlotWaveformFitting(self):
            savepath=save_tests()
            rfstream= rfMoho_example()
            VpSed=2.2
            HSed=0.1
            VpCrust = 6.9
            gaussian=1.25
            rayp=0.04
            KMoho= np.linspace(1.65,1.95,121)
            HMoho=np.linspace(20,60,201)
            # KMoho= np.linspace(1.65,1.95,5)
            # HMoho=np.linspace(20,60,5)
            wtCorr, wtRMSE, wtPG = [0.5, 0.3, 0.2] #wtCorr+wtRMSE+wtPG=1.0
            format='pdf'
            FittingResult=WaveformFitting(rfstream, HSed, VpSed, VpCrust, rayp, KMoho, HMoho, 
                        gaussian, wtCorr, wtRMSE, wtPG, savepath, format)
            PlotWaveformFitting(FittingResult, wtCorr, wtRMSE, wtPG, savepath, format)
    

def suite():
    return unittest.makeSuite(TestWaveformFitting, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')


#%% Test the function for HK stacking
import os 
import numpy as np
from rfsed.rf.rfstream import read_rf
from rfsed.hkZhu import hk, plothk
import h5py, h5
from obspy import read
from rfsed.util import rfMoho_example, rfSed_example, save_plot
staname='OPLO'
w1, w2, w3 = [0.6, 0.3, 0.1]
savedir=save_plot()
savepath = savedir + '/HK_Stacking_Zhu/'
if not os.path.exists(savepath):  # create data folder if necessary
    os.mkdir(savepath)
#----------------------------------------------------------
#%% Moho Example 
rfst = rfMoho_example()
rfstreams = rfst.select(component='R', station=staname)
K= np.linspace(1.65,1.95,121)
H=np.linspace(20,60,201)
Vp=6.9
Result=hk(rfstreams, layer='Moho', stack=True, w1 = w1, w2 = w2, w3 = w3, K= K, H=H, Vp = Vp)
# Result=hk(rfstreams, layer='Moho', stack=False, w1 = w1, w2 = w2, w3 = w3, K= K, H=H, Vp = Vp)
plothk(HKResult=Result, g = [75.,10., 15., 2.5], rmneg = None, savepath=savepath, format = 'jpg')
#----------------------------------------------------------
#%% Sediment Example
# rfst = rfSed_example()
# rfstreams = rfst.select(component='R', station=staname)
# K= np.linspace(1.65,2.25,201)
# H=np.linspace(0,10,201)
# Vp=2.5
# Result=hk(rfstreams, layer='Sed', stack=False, w1 = w1, w2 = w2, w3 = w3, K= K, H=H, Vp = Vp)
# # Result=hk(rfstreams, layer='Sed', stack=True, w1 = w1, w2 = w2, w3 = w3, K= K, H=H, Vp = Vp)
# plothk(HKResult=Result, g = [75.,10., 15., 2.5], rmneg = None, savepath=savepath, format = 'jpg')
# %%



"""
.. module:: hkYu
        :synopsis: Modified H-K stacking method of Yu et al. (2015) for receiver functions filtered with the resonance filter
        Yu, Y., Song, J., Liu, K. H., & Gao, S. S. (2015). Determining crustal structure
        beneath seismic stations overlying a low-velocity sedimentary layer using receiver
        functions. Journal of Geophysical Research: Solid Earth, 120 , 3208-3218. doi:
        10.1002/2014JB011610
.. moduleauthor:: Stephen Akinremi <s.akinremi@utwente.nl> Islam Fadel <i.e.a.m.fadel@utwente.nl> (October 2023)
"""

import numpy as np
from scipy import signal
from scipy.signal import argrelextrema, argrelmin, argrelmax
import math
import cmath
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.transforms as mtransforms
from scipy.fftpack import fft, ifft
from matplotlib.legend_handler import HandlerLine2D as HL
import seaborn
seaborn.set_style("darkgrid", {"axes.facecolor": ".85"})

#----------------------------------------------------------
def getamp(rfdata, tarray, t):
    """
    Get the amplitude of the receiver function at a specific time

    :param rfdata: receiver function data
    :type rfdata: numpy array
    :param tarray: time array
    :type tarray: numpy array
    :param t: time to get the amplitude
    :type t: float
    :return: amplitude of the receiver function at time t
    """
    amp = rfdata[(np.abs(tarray - t).argmin())]
    return amp
#----------------------------------------------------------
def hkYu(FltResult, rayp=0.04, HSubSed=np.linspace(20,60,201), KSubSed=np.linspace(1.65,1.95,121), 
               HSed=np.linspace(0,10,201), KSed=np.linspace(1.65,2.25,201), VpMoho=6.9, VpSed= 2.5,  
               w1SubSed=0.6, w2SubSed=0.3, w3SubSed=0.1, w1Sed=0.6, w2Sed=0.3, w3Sed=0.1):
    """
    Modified H-K stacking method of Yu et al. (2015)

    :param FltResult: Dictionary of results from the resonance filtering method (function: Resonance_Filt)
    :type FltResult: dict
    :param rayp: Ray parameter
    :type rayp: float
    :param HSubSed: Subsediment layer thickness array
    :type HSubSed: numpy array
    :param KSubSed: Subsediment layer Vp/Vs array
    :type KSubSed: numpy array
    :param HSed: Sediment layer thickness array
    :type HSed: numpy array
    :param KSed: Sediment layer Vp/Vs array
    :type KSed: numpy array
    :param VpMoho: Moho Vp
    :type VpMoho: float
    :param VpSed: Sediment Vp
    :type VpSed: float
    :param w1SubSed: Weight for the Subsediment Ps arrival at adjusted arrival time
    :type w1SubSed: float
    :param w2SubSed: Weight for the Subsediment Ppps arrival at adjusted arrival time
    :type w2SubSed: float
    :param w3SubSed: Weight for the Subsediment Psps+PpSs arrival at adjusted arrival time
    :type w3SubSed: float
    :param w1Sed: Weight for the Ps of the sediment layer
    :type w1Sed: float
    :param w2Sed: Weight for the Ppps of the Subsediment layer
    :type w2Sed: float
    :param w3Sed: Weight for the Psps+PpSs of the Subsediment layer
    :type w3Sed: float
    """
    if VpMoho is None: VpMoho = 6.5
    if VpSed is None: VpSed = 2.5
    if w1SubSed is None: w1 = 0.6
    if w2SubSed is None: w2 = 0.3
    if w3SubSed is None: w3 = 0.1   
    #----------------------------------------------------------
    rp=rayp
    Fltrf=FltResult['filteredrf']
    rfdata = Fltrf
    t = FltResult['time']
    tlag=FltResult['tlag']
    VsSed=0.7858 - 1.2344*VpSed + 0.7949*VpSed**2 - 0.1238*VpSed**3 + 0.0064*VpSed**4
    SedH=tlag*VsSed/2
    Pdelay=SedH/VsSed - SedH/VpSed
    #----------------------------------------------------------
    # Stack for SubSediment Layer
    stkSubSed = np.zeros((len(KSubSed)*len(HSubSed),3))
    z = 0 
    for i in range(len(KSubSed)):
        Ktemp = KSubSed[i]
        for j in range(len(HSubSed)):
            Htemp = HSubSed[j]
            s = 0.0
            trdata = rfdata
            #--------------------------------------
            term1= ((Ktemp/VpMoho)**2 - rp**2)**0.5
            term2= ((1/VpMoho)**2 - rp**2)**0.5
            tps = Htemp * (term1 - term2)
            tppps = Htemp * (term1 + term2)
            tpsps = Htemp * 2 * (term1)
            #--------------------------------------
            # Adjusted Arrival Times for subsediment layer stacking
            tpsSubSed=tps+Pdelay
            tpppsSubSed=tppps+tlag - Pdelay
            tpspsSubSed=tpsps+tlag
            #--------------------------------------
            ampps = getamp(trdata, t, tpsSubSed)
            ampppps = getamp(trdata,t, tpppsSubSed)
            amppsps = getamp(trdata, t, tpspsSubSed)
            #--------------------------------------
            stemp = (w1SubSed * ampps) + (w2SubSed * ampppps) - (w3SubSed * amppsps)
            s = stemp + s
            stkSubSed[z,:] = [Htemp, Ktemp, s]
            z = z + 1
    bmodSubSed = stkSubSed[np.argmax(stkSubSed[:,2]),:]
    #Show the best model
    plt.tricontourf(stkSubSed[:,0], stkSubSed[:,1], stkSubSed[:,2],60, cmap='jet')
    cb=plt.colorbar(format='%.3f', orientation="vertical")
    p1,=plt.plot(bmodSubSed[0],bmodSubSed[1], 'k+', mew=5, ms=15,\
        label='Best Model %s km %s Vp/Vs'%(bmodSubSed[0], bmodSubSed[1]))
    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,\
        ncol=2, mode="expand", borderaxespad=0, \
        handler_map={p1:HL(numpoints=1)}) 
    plt.ylabel('Vp/Vs')
    plt.xlabel('Depth km')
    plt.xlim(min(HSubSed),max(HSubSed))
    plt.ylim(min(KSubSed),max(KSubSed))
    # plt.savefig(savepath , format=format, dpi=250) 
    plt.show()
    plt.close("all")  
    HSubSedbm = bmodSubSed[0]
    KSubSedbm = bmodSubSed[1]
    SSubSedbm = bmodSubSed[2]
    print("Best Subsediment thickness: ", HSubSedbm, "Best Subsediment Vp/Vs:", KSubSedbm, "Max stack: ", SSubSedbm)
    #----------------------------------------------------------
    # Stack for Sediment Layer
    stkSed = np.zeros((len(KSed)*len(HSed),3))
    z = 0 
    for i in range(len(KSed)):
        Ktemp = KSed[i]
        for j in range(len(HSed)):
            Htemp = HSed[j]
            s = 0.0
            trdata = rfdata
            #--------------------------------------
            # Calculate tps for Sediment phase
            term3= ((Ktemp/VpSed)**2 - rp**2)**0.5
            term4= ((1/VpSed)**2 - rp**2)**0.5
            tpsSed = Htemp * (term3 - term4)
            #--------------------------------------
            # tpps and tpsps are fixed based on the best subsediment model
            term5= ((KSubSedbm /VpMoho)**2 - rp**2)**0.5
            term6= ((1/VpMoho)**2 - rp**2)**0.5
            tpppsSubSed = HSubSedbm * (term5 + term6)
            tpspsSubSed = HSubSedbm * 2 * (term5)  
            #--------------------------------------
            ampps = getamp(trdata, t, tpsSed)
            ampppps = getamp(trdata,t, tpppsSubSed)
            amppsps = getamp(trdata, t, tpspsSubSed)
            #--------------------------------------
            stemp = (w1SubSed * ampps) + (w2SubSed * ampppps) - (w3SubSed * amppsps)
            s = stemp + s
            stkSed[z,:] = [Htemp, Ktemp, s]
            z = z + 1
    bmodSed = stkSed[np.argmax(stkSed[:,2]),:]
    #Show the best model
    plt.tricontourf(stkSed[:,0], stkSed[:,1], stkSed[:,2],60, cmap='jet')
    cb=plt.colorbar(format='%.3f', orientation="vertical")
    p1,=plt.plot(bmodSed[0],bmodSed[1], 'k+', mew=5, ms=15,\
        label='Best Model %s km %s Vp/Vs'%(bmodSed[0], bmodSed[1]))
    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,\
        ncol=2, mode="expand", borderaxespad=0, \
        handler_map={p1:HL(numpoints=1)}) 
    plt.ylabel('Vp/Vs')
    plt.xlabel('Depth km')
    plt.xlim(min(HSed),max(HSed))
    plt.ylim(min(KSed),max(KSed))
    # plt.savefig(savepath , format=format, dpi=250) 
    plt.show()
    plt.close("all")  
    HSedbm = bmodSed[0]
    KSedbm = bmodSed[1]
    SSedbm = bmodSed[2]
    print("Best Sediment thickness: ", HSedbm, " km", "Best Sediment Vp/Vs:", KSedbm, "Max stack: ", SSedbm)
    print("Best Subsediment thickness: ", HSubSedbm, " km", "Best Subsediment Vp/Vs:", KSubSedbm, "Max stack: ", SSubSedbm)
    HMohobm=HSubSedbm + HSedbm
    print("Best Moho thickness: ", HMohobm, " km")
    HKYuResult = {'FltRF': Fltrf, 'time':t, 'tlag':tlag, 'Pdelay':Pdelay, 'rayp':rayp, 'BestMohoDepth': HMohobm, 
                  'bestmodelSubSed':bmodSubSed, 'VpMoho':VpMoho, 'VpSed':VpSed, 'stackvaluesSubSed':stkSubSed, 
                  'bestmodelSed':bmodSed, 'stackvaluesSed':stkSed,'HMoho':HSubSed, 'KMoho':KSubSed,'HSed':HSed, 
                  'KSed':KSed,'SubSedThick':HSubSedbm, 'SubSedVpVs':KSubSedbm, 'SedThick':HSedbm, 'SedVpVs':KSedbm}
    return HKYuResult
#----------------------------------------------------------
def plothkYu(hkYuResult, savepath, g = [75.,10., 15., 2.5], rmneg = None, format = 'jpg'): 
    """
    Plot the results of the Modified H-K stacking method of Yu et al. (2015)

    :param hkYuResult: Dictionary of results from the Modified H-K stacking method (function: hkYu)
    :type hkYuResult: dict
    :param savepath: path to save the plots
    :type savepath: str
    :param g: gain values for the plot (default: [75.,10., 15., 2.5])
    :type g: list
    :param rmneg: remove the stack values that is lower than zero (default: False)
    :type rmneg: bool
    :param format: format of the plot (default: jpg)
    :type format: str   
    """
    if rmneg is None: rmneg = False
    if format is None: format = 'pdf'
    if g is None: g = [75.,10., 15., 2.5]
    staname='FilteredRF_HKYu'
    Pdelay=hkYuResult['Pdelay']
    tlag=hkYuResult['tlag']
    VpMoho=hkYuResult['VpMoho']
    VpSed=hkYuResult['VpSed']
    HSed=hkYuResult['HSed']
    KSed=hkYuResult['KSed']
    HMoho=hkYuResult['HMoho']
    KMoho=hkYuResult['KMoho']
    bmodSed = hkYuResult['bestmodelSed']
    H1 = bmodSed[0]
    K1 = bmodSed[1]
    stkSed=hkYuResult['stackvaluesSed']
    bmodSubSed = hkYuResult['bestmodelSubSed']
    H2 = bmodSubSed[0]
    K2 = bmodSubSed[1]
    stkSubSed=hkYuResult['stackvaluesSubSed']
    #----------------------------------------------------------
    # Plot for Sediment Layer
    rfdata=hkYuResult['FltRF']
    rp = hkYuResult['rayp']
    t=hkYuResult['time']
    #--------------------------------------
    # Calculate tps for Sediment phase
    term3= ((K1/VpSed)**2 - rp**2)**0.5
    term4= ((1/VpSed)**2 - rp**2)**0.5
    tpsSed = H1 * (term3 - term4)
    #--------------------------------------
    # tpps and tpsps are fixed based on the best subsediment model
    term5= ((K2 /VpMoho)**2 - rp**2)**0.5
    term6= ((1/VpMoho)**2 - rp**2)**0.5
    tpppsSubSed = H2 * (term5 + term6)
    tpspsSubSed = H2 * 2 * (term5)  
    #----------------------------------------------------------
    # plot H-K results
    #----------------------------------------------------------
    plt.figure(figsize=(18, 6)) 
    gs = gridspec.GridSpec(1, 3, width_ratios=[6,4,4])
    #-----Remove the stack values that is lower than zero------
    if rmneg == True:
        for i in range(len(stkSed)):
            if stkSed[i,2] <= 0:
                stkSed[i,2] = 0
    #----------------------------------------------------------
    ax1 = plt.subplot(gs[0, 0])
    plt.tricontourf(stkSed[:,0], stkSed[:,1], stkSed[:,2],50, cmap ="jet")
    #----------------------------------------------------------
    cb=plt.colorbar(format='%.3f', orientation="vertical")
    p1,=ax1.plot(bmodSed[0],bmodSed[1], 'k+', mew=2, ms=10, \
    label='Best Model Sediment %.2f km %.2f Vp/Vs'%(bmodSed[0],bmodSed[1]+0.001))
    props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,ncol=2, mode="expand", borderaxespad=0, handler_map={p1:HL(numpoints=1)})
    ax1.set_ylabel('Vp/Vs')
    ax1.set_xlabel('Depth km')
    ax1.set_xlim(min(HSed), max(HSed))
    ax1.set_ylim(min(KSed), max(KSed))
    #----------------------------------------------------------
    # Plot the receiver function witht the estimated times
    #----------------------------------------------------------
    ax2 = plt.subplot(gs[0, 1]) 
    baz=180
    rfdata = ((rfdata) * (g[0]/2)) + baz
    major_ticks_x = np.arange(-10, 41, 5)                                              
    minor_ticks_x = np.arange(-10, 41, 5) 
    major_ticks_y = np.arange(0, 361, 45)                                              
    minor_ticks_y = np.arange(-30, 390, 5)
    ax2.set_xticks(major_ticks_x)                                                       
    ax2.set_xticks(minor_ticks_x, minor=True)                                           
    ax2.set_yticks(major_ticks_y)                                                       
    ax2.set_yticks(minor_ticks_y, minor=True)                                          
    ax2.set_xlim(-5, 30)
    ax2.set_ylim(-29, 390)
    ax2.set_xlabel("Time (sec)")
    ax2.set_ylabel("Back-Azimuth (deg)")  
    ax2.plot(t, rfdata, "k-", lw=0.5)
    ax2.fill_between(t, baz, rfdata, where=rfdata > baz, facecolor='red', alpha = 0.25)
    ax2.fill_between(t, baz, rfdata, where=rfdata <= baz, facecolor='blue', alpha = 0.25)
    mtransforms.blended_transform_factory(ax2.transData, ax2.transAxes)
    ampgain = g[1]
    ax2.plot([tpsSed, tpsSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5, label="Moho")
    ax2.plot([tpppsSubSed, tpppsSubSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
    ax2.plot([tpspsSubSed, tpspsSubSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
    filename = "%s/%s_%s.%s"%(savepath, staname, "Sediment", format)
    plt.savefig(filename , format=format, transparent=False,\
        dpi=250, bbox_inches = 'tight', pad_inches=0.1)
    plt.show()
    plt.close('all')
    #----------------------------------------------------------
    # Plot SubSediment Stack
    #----------------------------------------------------------
    rfdata=hkYuResult['FltRF']
    bmodSubSed = hkYuResult['bestmodelSubSed']
    H2 = bmodSubSed[0]
    K2 = bmodSubSed[1]
    stkSubSed=hkYuResult['stackvaluesSubSed']
    term1= ((K2/VpMoho)**2 - rp**2)**0.5
    term2= ((1/VpMoho)**2 - rp**2)**0.5
    tps = H2 * (term1 - term2)
    tppps = H2 * (term1 + term2)
    tpsps = H2 * 2 * (term1)
    #--------------------------------------
    # Adjusted Arrival Times for subsediment layer stacking
    tpsSubSed=tps+Pdelay
    tpppsSubSed=tppps+tlag - Pdelay
    tpspsSubSed=tpsps+tlag
    #----------------------------------------------------------
    plt.figure(figsize=(18, 6)) 
    gs = gridspec.GridSpec(1, 3, width_ratios=[6,4,4])
    #-----Remove the stack values that is lower than zero------
    if rmneg == True:
        for i in range(len(stkSubSed)):
            if stkSubSed[i,2] <= 0:
                stkSubSed[i,2] = 0
    #----------------------------------------------------------
    ax1 = plt.subplot(gs[0, 0])
    plt.tricontourf(stkSubSed[:,0], stkSubSed[:,1], stkSubSed[:,2],50, cmap ="jet")
    cb=plt.colorbar(format='%.3f', orientation="vertical")
    cb.ax.set_xlabel('S', rotation=0)
    p1,=ax1.plot(bmodSubSed[0],bmodSubSed[1], 'k+', mew=2, ms=10, \
    label='Best SubSediment Model %.2f km %.2f Vp/Vs'%(bmodSubSed[0],bmodSubSed[1]+0.001))
    props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,\
        ncol=2, mode="expand", borderaxespad=0, \
        handler_map={p1:HL(numpoints=1)})
    ax1.set_ylabel('Vp/Vs')
    ax1.set_xlabel('Depth km')
    ax1.set_xlim(min(HMoho), max(HMoho))
    ax1.set_ylim(min(KMoho), max(KMoho))
    #----------------------------------------------------------
    # Plot the receiver function witht the estimated times
    #----------------------------------------------------------
    ax2 = plt.subplot(gs[0, 1]) 
    baz=180
    rfdata = ((rfdata) * g[0]) + baz
    major_ticks_x = np.arange(-10, 41, 5)                                           
    minor_ticks_x = np.arange(-10, 41, 1) 
    major_ticks_y = np.arange(0, 361, 45)                                              
    minor_ticks_y = np.arange(-30, 390, 5)
    ax2.set_xticks(major_ticks_x)                                                       
    ax2.set_xticks(minor_ticks_x, minor=True)                                           
    ax2.set_yticks(major_ticks_y)                                                       
    ax2.set_yticks(minor_ticks_y, minor=True)                                  
    ax2.set_xlim(-5, 30)
    ax2.set_ylim(-29, 390)
    ax2.plot(t, rfdata, "k-", lw=0.5)
    ax2.fill_between(t, baz, rfdata, where=rfdata > baz, 
                        facecolor='red', alpha = 0.25)
    ax2.fill_between(t, baz, rfdata, where=rfdata <= baz, 
                        facecolor='blue', alpha = 0.25)
    mtransforms.blended_transform_factory(ax2.transData, ax2.transAxes)
    #----------------------------------------------------------
    # Plot H-K sequential formula times
    #----------------------------------------------------------
    ampgain = g[1]
    ax2.plot([tpsSubSed, tpsSubSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5, label="Moho")
    ax2.plot([tpppsSubSed, tpppsSubSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
    ax2.plot([tpspsSubSed, tpspsSubSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
    filename = "%s/%s_%s.%s"%(savepath, staname, 'SubSediment', format)
    plt.savefig(filename , format=format, transparent=False,\
    dpi=250, bbox_inches = 'tight', pad_inches=0.1)
    plt.show()
    plt.close('all')
#----------------------------------------------------------



#%% Test the function Sequential Stacking for H-K
import os 
import numpy as np
from rfsed.rf.rfstream import read_rf
from rfsed.hkSeqYeck import hkSeq, plotSeqhk
from rfsed.util import rfMoho_example, rfSed_example, save_plot

savedir=save_plot()
savepath = savedir + '/HK_Sequential_Yeck/'
if not os.path.exists(savepath):  # create data folder if necessary
    os.mkdir(savepath)
staname='OPLO'
#----------------------------------------------------------
# Sediment Parameters
rfstSed = rfSed_example()
rfstreamSed = rfstSed.select(component='R', station=staname)
KSed= np.linspace(1.65,2.25,201)
HSed=np.linspace(0,10,201)
VpSed=2.5
w1Sed, w2Sed, w3Sed = [0.6, 0.3, 0.1]
#----------------------------------------------------------
# Moho Parameters
rfstMoho = rfMoho_example()
rfstreamMoho = rfstMoho.select(component='R', station=staname)
KMoho= np.linspace(1.65,1.95,121)
HMoho=np.linspace(20,60,201)
VpMoho=6.9
w1Moho, w2Moho, w3Moho = [0.6, 0.3, 0.1]
#----------------------------------------------------------
SequentialHKResult = hkSeq(rfstreamSed, rfstreamMoho,  w1Sed = w1Sed, 
             w2Sed = w2Sed, w3Sed=w3Sed, KSed=KSed, HSed=HSed, VpSed=VpSed,
             w1Moho = w1Moho, w2Moho = w2Moho, w3Moho=w3Moho, KMoho=KMoho, 
             HMoho=HMoho, VpMoho=VpMoho, stack = False)
plotSeqhk(SequentialHKResult=SequentialHKResult, g = [75.,10., 15., 2.5], 
          rmneg = None, savepath=savepath, format = 'jpg')
#%% Stacked RF Example
# SequentialHKResult = hkSeq(rfstreamSed, rfstreamMoho,  w1Sed = w1Sed, 
#              w2Sed = w2Sed, w3Sed=w3Sed, KSed=KSed, HSed=HSed, VpSed=VpSed,
#              w1Moho = w1Moho, w2Moho = w2Moho, w3Moho=w3Moho, KMoho=KMoho, 
#              HMoho=HMoho, VpMoho=VpMoho, stack = True)
# plotSeqhk(SequentialHKResult=SequentialHKResult, g = [75.,10., 15., 2.5], 
#           rmneg = None, savepath=savepath, format = 'jpg')


# Copyright 2013-2019 Tom Eulenfeld, MIT license
"""
All objects and functions below are after the Tom Eulenfeld's
rf project License below.

The MIT License (MIT)

Copyright (c) 2013-2019 Tom Eulenfeld

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
"""
Classes and functions for receiver function calculation.
"""
import json
from operator import itemgetter
from pkg_resources import resource_filename
import warnings

import numpy as np
from obspy import read, Stream, Trace
from obspy.core import AttribDict
from obspy.geodetics import gps2dist_azimuth
from obspy.taup import TauPyModel
from rfsed.rf.deconvolve import deconvolve
from rfsed.rf.simple_model import load_model
from rfsed.rf.util import DEG2KM, IterMultipleComponents, _add_processing_info


def __get_event_origin_prop(h):
    def wrapper(event):
        try:
            r = (event.preferred_origin() or event.origins[0])[h]
        except IndexError:
            raise ValueError('No origin')
        if r is None:
            raise ValueError('No origin ' + h)
        if h == 'depth':
            r = r / 1000
        return r
    return wrapper


def __get_event_magnitude(event):
    try:
        return (event.preferred_magnitude() or event.magnitudes[0])['mag']
    except IndexError:
        raise ValueError('No magnitude')


def __get_event_id(event):
    evid = event.get('resource_id')
    if evid is not None:
        evid = str(evid)
    return evid


def __SAC2UTC(stats, head):
    from obspy.io.sac.util import get_sac_reftime
    return get_sac_reftime(stats.sac) + stats[head]


def __UTC2SAC(stats, head):
    from obspy.io.sac.util import get_sac_reftime
    return stats[head] - get_sac_reftime(stats.sac)


_STATION_GETTER = (('station_latitude', itemgetter('latitude')),
                   ('station_longitude', itemgetter('longitude')),
                   ('station_elevation', itemgetter('elevation')))
_EVENT_GETTER = (
    ('event_latitude', __get_event_origin_prop('latitude')),
    ('event_longitude', __get_event_origin_prop('longitude')),
    ('event_depth', __get_event_origin_prop('depth')),
    ('event_magnitude', __get_event_magnitude),
    ('event_time', __get_event_origin_prop('time')),
    ('event_id', __get_event_id))

# header values which will be written to waveform formats (SAC and Q)
# H5 simply writes all stats entries
_HEADERS = (tuple(zip(*_STATION_GETTER))[0] +
            tuple(zip(*_EVENT_GETTER))[0][:-1] + (  # do not write event_id
            'onset', 'type', 'phase', 'moveout',
            'distance', 'back_azimuth', 'inclination', 'slowness',
            'pp_latitude', 'pp_longitude', 'pp_depth',
            'box_pos', 'box_length'))

# The corresponding header fields in the format
# The following headers can at the moment only be stored for H5:
# slowness_before_moveout, box_lonlat, event_id
_FORMATHEADERS = {'sac': ('stla', 'stlo', 'stel', 'evla', 'evlo',
                          'evdp', 'mag', 'o', 'a',
                          'kuser0', 'kuser1', 'kuser2',
                          'gcarc', 'baz', 'user0', 'user1',
                          'user2', 'user3', 'user4',
                          'user5', 'user6'),
                  # field 'COMMENT' is violated for different information
                  'sh': ('COMMENT', 'COMMENT', 'COMMENT',
                         'LAT', 'LON', 'DEPTH',
                         'MAGNITUDE', 'ORIGIN', 'P-ONSET',
                         'COMMENT', 'COMMENT', 'COMMENT',
                         'DISTANCE', 'AZIMUTH', 'INCI', 'SLOWNESS',
                         'COMMENT', 'COMMENT', 'COMMENT',
                         'COMMENT', 'COMMENT')}
_HEADER_CONVERSIONS = {'sac': {'onset': (__SAC2UTC, __UTC2SAC),
                               'event_time': (__SAC2UTC, __UTC2SAC)}}


_TF = '.datetime:%Y-%m-%dT%H:%M:%S'

_H5INDEX = {
    'rf': ('waveforms/{network}.{station}.{location}/{event_time%s}/' % _TF +
           '{channel}_{starttime%s}_{endtime%s}' % (_TF, _TF)),
    'profile': 'waveforms/{channel[2]}_{box_pos}'
}


def read_rf(pathname_or_url=None, format=None, **kwargs):
    """
    Read waveform files into RFStream object.

    See :func:`~obspy.core.stream.read` in ObsPy.
    """
    if pathname_or_url is None:   # use example file
        fname = resource_filename('rfsed', 'rf/example/minimal_example.tar.gz')
        pathname_or_url = fname
        format = 'SAC'
    stream = read(pathname_or_url, format=format, **kwargs)
    return RFStream(stream)


class RFStream(Stream):

    """
    Class providing the necessary functions for receiver function calculation.

    :param traces: list of traces, single trace or stream object

    To initialize a RFStream from a Stream object use

    >>> rfstream = RFStream(stream)

    To initialize a RFStream from a file use

    >>> rfstream = read_rf('test.SAC')

    Format specific headers are loaded into the stats object of all traces.
    """

    def __init__(self, traces=None):
        self.traces = []
        if isinstance(traces, Trace):
            traces = [traces]
        if traces:
            for tr in traces:
                if not isinstance(tr, RFTrace):
                    tr = RFTrace(trace=tr)
                self.traces.append(tr)

    def __is_set(self, header):
        return all(header in tr.stats for tr in self)

    def __get_unique_header(self, header):
        values = set(tr.stats[header] for tr in self if header in tr.stats)
        if len(values) > 1:
            warnings.warn('Header %s has different values in stream.' % header)
        if len(values) == 1:
            return values.pop()

    @property
    def type(self):
        """Property for the type of stream, 'rf', 'profile' or None"""
        return self.__get_unique_header('type')

    @type.setter
    def type(self, value):
        for tr in self:
            tr.stats.type = value

    @property
    def method(self):
        """Property for used rf method, 'P' or 'S'"""
        phase = self.__get_unique_header('phase')
        if phase is not None:
            return phase[-1].upper()

    @method.setter
    def method(self, value):
        for tr in self:
            tr.stats.phase = value

    def write(self, filename, format, sh_compat=False, **kwargs):
        """
        Save stream to file including format specific headers.

        See `Stream.write() <obspy.core.stream.Stream.write>` in ObsPy.

        :param sh_compat: Ensure files in Q format can be read with
            SeismicHandler (default: False). If set to True the COMMENT field
            will be deleted which might result in loss of some metadata.
            (see issue #37).
        """
        if len(self) == 0:
            return
        for tr in self:
            tr._write_format_specific_header(format, sh_compat=sh_compat)
            if format.upper() == 'Q':
                tr.stats.station = tr.id
        if format.upper() == 'H5':
            index = self.type
            if index is None and 'event_time' in self[0].stats:
                index = 'rf'
            if index:
                import obspyh5
                old_index = obspyh5._INDEX
                obspyh5.set_index(_H5INDEX[index])
        super(RFStream, self).write(filename, format, **kwargs)
        if format.upper() == 'H5' and index:
            obspyh5.set_index(old_index)
        if format.upper() == 'Q':
            for tr in self:
                tr.stats.station = tr.stats.station.split('.')[1]

    def trim2(self, starttime=None, endtime=None, reftime=None, **kwargs):
        """
        Alternative trim method accepting relative times.

        See :meth:`~obspy.core.stream.Stream.trim`.

        :param starttime,endtime: accept UTCDateTime or seconds relative to
            reftime
        :param reftime: reference time, can be an UTCDateTime object or a
            string. The string will be looked up in the stats dictionary
            (e.g. 'starttime', 'endtime', 'onset').
        """
        for tr in self.traces:
            t1 = tr._seconds2utc(starttime, reftime=reftime)
            t2 = tr._seconds2utc(endtime, reftime=reftime)
            tr.trim(t1, t2, **kwargs)
        self.traces = [_i for _i in self.traces if _i.stats.npts]
        return self

    def slice2(self, starttime=None, endtime=None, reftime=None,
               keep_empty_traces=False, **kwargs):
        """
        Alternative slice method accepting relative times.

        See :meth:`~obspy.core.stream.Stream.slice` and `trim2()`.
        """
        traces = []
        for tr in self:
            t1 = tr._seconds2utc(starttime, reftime=reftime)
            t2 = tr._seconds2utc(endtime, reftime=reftime)
            sliced_trace = tr.slice(t1, t2, **kwargs)
            if not keep_empty_traces and not sliced_trace.stats.npts:
                continue
            traces.append(sliced_trace)
        return self.__class__(traces)

    def deconvolve(self, *args, **kwargs):
        """
        Deconvolve source component of stream.

        All args and kwargs are passed to the function
        `~rf.deconvolve.deconvolve()`.
        """
        rsp = deconvolve(self, *args, **kwargs)
        self.traces = rsp.traces
        return self

    @_add_processing_info
    def rf(self, method=None, filter=None, trim=None, downsample=None,
           rotate='ZNE->LQT', deconvolve='time', source_components=None,
           **kwargs):
        r"""
        Calculate receiver functions in-place.

        :param method: 'P' for P receiver functions, 'S' for S receiver
            functions, if None method will be determined from the phase
        :param dict filter: filter stream with its
            `~obspy.core.stream.Stream.filter` method and given kwargs
        :type trim: tuple (start, end)
        :param trim: trim stream relative to P- or S-onset
             with `trim2()` (seconds)
        :param float downsample: downsample stream with its
            :meth:`~obspy.core.stream.Stream.decimate` method to the given
            frequency
        :type rotate: 'ZNE->LQT' or 'NE->RT'
        :param rotate: rotate stream with its
            :meth:`~obspy.core.stream.Stream.rotate`
            method with the angles given by the back_azimuth and inclination
            attributes of the traces stats objects. You can set these to your
            needs or let them be computed by :func:`~rf.rfstream.rfstats`.
        :param deconvolve: 'time', 'waterlevel', 'iterative' or 'multitaper'
            for time domain damped, frequency domain water level,
            time domain iterative, or frequency domain multitaper
            deconvolution using the stream's `deconvolve()` method.
            See `~.deconvolve.deconvolve()`,
            `.deconv_time()`, `.deconv_waterlevel()`,
            `.deconv_iterative()`, and `.deconv_multitaper()`
            for further documentation.
        :param source_components: parameter is passed to deconvolve.
            If None, source components will be chosen depending on method.
        :param \*\*kwargs: all other kwargs not mentioned here are
            passed to deconvolve

        After performing the deconvolution the Q/R and T components are
        multiplied by -1 to get a positive phase for a Moho-like positive
        velocity contrast. Furthermore for method='S' all components are
        mirrored at t=0 for a better comparison with P receiver functions.
        See source code of this function for the default
        deconvolution windows.
        """
        def iter3c(stream):
            return IterMultipleComponents(stream, key='onset',
                                          number_components=(2, 3))

        if method is None:
            method = self.method
        if method is None or method not in 'PS':
            msg = "method must be one of 'P', 'S', but is '%s'"
            raise ValueError(msg % method)
        if source_components is None:
            source_components = 'LZ' if method == 'P' else 'QR'
        if filter:
            self.filter(**filter)
        if trim:
            self.trim2(*trim, reftime='onset')
        if downsample:
            for tr in self:
                if downsample <= tr.stats.sampling_rate:
                    tr.decimate(int(tr.stats.sampling_rate) // downsample)
        if rotate:
            for stream3c in iter3c(self):
                stream3c.rotate(rotate)
        # Multiply -1 on Q component, because Q component is pointing
        # towards the event after the rotation with ObsPy.
        # For a positive phase at a Moho-like velocity contrast,
        # the Q component has to point away from the event.
        # This is not necessary for the R component which points already
        # away from the event.
        # (compare issue #4)
        for tr in self:
            if tr.stats.channel.endswith('Q'):
                tr.data = -tr.data
        if deconvolve:
            for stream3c in iter3c(self):
                kwargs.setdefault('winsrc', method)
                stream3c.deconvolve(method=deconvolve,
                                    source_components=source_components,
                                    **kwargs)
        # Mirrow Q/R and T component at 0s for S-receiver method for a better
        # comparison with P-receiver method (converted Sp wave arrives before
        # S wave, but converted Ps wave arrives after P wave)
        if method == 'S':
            for tr in self:
                tr.data = tr.data[::-1]
                tr.stats.onset = tr.stats.starttime + (tr.stats.endtime -
                                                       tr.stats.onset)
        self.type = 'rf'
        if self.method != method:
            self.method = method
        return self

    @_add_processing_info
    def moveout(self, phase=None, ref=6.4, model='iasp91'):
        """
        In-place moveout correction to a reference slowness.

        Needs stats attributes slowness and onset.

        :param phase: 'Ps', 'Sp', 'Ppss' or other multiples, if None is set to
            'Ps' for P receiver functions or 'Sp' for S receiver functions
        :param ref: reference ray parameter in s/deg
        :param model: Path to model file
            (see `.SimpleModel`, default: iasp91)
        """
        if phase is None:
            phase = self.method + {'P': 's', 'S': 'p'}[self.method]
        model = load_model(model)
        model.moveout(self, phase=phase, ref=ref)
        for tr in self:
            tr.stats.moveout = phase
            tr.stats.slowness_before_moveout = tr.stats.slowness
            tr.stats.slowness = ref
        return self

    def ppoints(self, pp_depth, pp_phase=None, model='iasp91'):
        """
        Return coordinates of piercing point calculated by 1D ray tracing.

        Piercing point coordinates are stored in the
        stats attributes plat and plon. Needs stats attributes
        station_latitude, station_longitude, slowness and back_azimuth.

        :param pp_depth: depth of interface in km
        :param pp_phase: 'P' for piercing points of P wave, 'S' for piercing
            points of S wave or multiples, if None will be
            set to 'S' for P receiver functions or 'P' for S receiver functions
        :param model: path to model file (see `.SimpleModel`, default: iasp91)
        :return: NumPy array with coordinates of piercing points
        """
        if pp_phase is None:
            pp_phase = {'P': 'S', 'S': 'P'}[self.method]
        model = load_model(model)
        for tr in self:
            model.ppoint(tr.stats, pp_depth, phase=pp_phase)
        return np.array([(tr.stats.pp_latitude, tr.stats.pp_longitude)
                         for tr in self])

    @_add_processing_info
    def stack(self):
        """
        Return stack of traces with same seed ids.

        Traces with same id need to have the same number of datapoints.
        Each trace in the returned stream will correspond to one unique seed
        id.
        """
        ids = set(tr.id for tr in self)
        tr = self[0]
        traces = []
        for id in ids:
            net, sta, loc, cha = id.split('.')
            data = np.mean([tr.data for tr in self if tr.id == id], axis=0)
            header = {'network': net, 'station': sta, 'location': loc,
                      'channel': cha, 'sampling_rate': tr.stats.sampling_rate}
            for entry in ('phase', 'moveout', 'station_latitude',
                          'station_longitude', 'station_elevation',
                          'processing'):
                if entry in tr.stats:
                    header[entry] = tr.stats[entry]
            tr2 = RFTrace(data=data, header=header)
            if 'onset' in tr.stats:
                onset = tr.stats.onset - tr.stats.starttime
                tr2.stats.onset = tr2.stats.starttime + onset
            traces.append(tr2)
        return self.__class__(traces)

    def profile(self, *args, **kwargs):
        """
        Return profile of receiver functions in the stream.

        See `.profile.profile()` for help on arguments.
        """
        from rfsed.rf.profile import profile
        return profile(self, *args, **kwargs)

    def plot_rf(self, *args, **kwargs):
        """
        Create receiver function plot.

        See `.imaging.plot_rf()` for help on arguments.
        """
        from rfsed.rf.imaging import plot_rf
        return plot_rf(self, *args, **kwargs)

    def plot_profile(self, *args, **kwargs):
        """
        Create receiver function profile plot.

        See `.imaging.plot_profile()` for help on arguments.
        """
        from rfsed.rf.imaging import plot_profile
        return plot_profile(self, *args, **kwargs)


class RFTrace(Trace):

    """
    Class providing the Trace object for receiver function calculation.
    """

    def __init__(self, data=np.array([]), header=None, trace=None):
        if header is None:
            header = {}
        if trace is not None:
            data = trace.data
            header = trace.stats
        super(RFTrace, self).__init__(data=data, header=header)
        st = self.stats
        if ('_format'in st and st._format.upper() == 'Q' and
                st.station.count('.') > 1):
            st.network, st.station, st.location = st.station.split('.')[:3]
        self._read_format_specific_header()

    def __str__(self, id_length=None):
        if 'onset' not in self.stats:
            return super(RFTrace, self).__str__(id_length=id_length)
        out = []
        type_ = self.stats.get('type')
        if type_ is not None:
            m = self.stats.get('phase')
            m = m[-1].upper() if m is not None else ''
            o1 = m + 'rf'
            if type_ != 'rf':
                o1 = o1 + ' ' + type_
            if self.id.startswith('...'):
                o1 = o1 + ' (%s)' % self.id[-1]
            else:
                o1 = o1 + ' ' + self.id
        else:
            o1 = self.id
        out.append(o1)
        t1 = self.stats.starttime - self.stats.onset
        t2 = self.stats.endtime - self.stats.onset
        o2 = '%.1fs - %.1fs' % (t1, t2)
        if self.stats.starttime.timestamp != 0:
            o2 = o2 + ' onset:%s' % self.stats.onset
        out.append(o2)
        out.append('{sampling_rate} Hz, {npts} samples')
        o3 = []
        if 'event_magnitude' in self.stats:
            o3.append('mag:{event_magnitude:.1f}')
        if 'distance' in self.stats:
            o3.append('dist:{distance:.1f}')
        if'back_azimuth' in self.stats:
            o3.append('baz:{back_azimuth:.1f}')
        if 'box_pos' in self.stats:
            o3.append('pos:{box_pos:.2f}km')
        if 'slowness' in self.stats:
            o3.append('slow:{slowness:.2f}')
        if 'moveout' in self.stats:
            o3.append('({moveout} moveout)')
        if np.ma.count_masked(self.data):
            o3.append('(masked)')
        out.append(' '.join(o3))
        return ' | '.join(out).format(**self.stats)

    def _read_format_specific_header(self, format=None):
        st = self.stats
        if format is None:
            if '_format' not in st:
                return
            format = st._format
        format = format.lower()
        if format == 'q':
            format = 'sh'
        try:
            header_map = zip(_HEADERS, _FORMATHEADERS[format])
        except KeyError:
            # file format is H5 or not supported
            return
        for head, head_format in header_map:
            try:
                value = st[format][head_format]
            except KeyError:
                continue
            else:
                if format == 'sac' and '-12345' in str(value):
                    pass
                elif format == 'sh' and head_format == 'COMMENT':
                    try:
                        st.update(json.loads(value))
                    except json.JSONDecodeError:
                        pass
                    continue
                else:
                    st[head] = value
            try:
                convert = _HEADER_CONVERSIONS[format][head][0]
                st[head] = convert(st, head)
            except KeyError:
                pass

    def _write_format_specific_header(self, format, sh_compat=False):
        st = self.stats
        format = format.lower()
        if format == 'q':
            format = 'sh'
        elif format == 'sac':
            # make sure SAC reference time is set
            from obspy.io.sac.util import obspy_to_sac_header
            self.stats.sac = obspy_to_sac_header(self.stats)
        try:
            header_map = zip(_HEADERS, _FORMATHEADERS[format])
        except KeyError:
            # file format is H5 or not supported
            return
        if format not in st:
            st[format] = AttribDict({})
        if format == 'sh':
            comment = {}
        for head, head_format in header_map:
            if format == 'sh' and head_format == 'COMMENT':
                try:
                    comment[head] = st[head]
                except KeyError:
                    pass
                continue
            try:
                val = st[head]
            except KeyError:
                continue
            try:
                convert = _HEADER_CONVERSIONS[format][head][1]
                val = convert(st, head)
            except KeyError:
                pass
            st[format][head_format] = val
        if format == 'sh':
            if sh_compat:
                st[format].pop('COMMENT', None)
            elif len(comment) > 0:
                def default(obj):  # convert numpy types
                    return np.asarray(obj).item()
                st[format]['COMMENT'] = json.dumps(
                    comment, separators=(',', ':'), default=default)

    def _seconds2utc(self, seconds, reftime=None):
        """Return UTCDateTime given as seconds relative to reftime"""
        from collections.abc import Iterable
        from obspy import UTCDateTime as UTC
        if isinstance(seconds, Iterable):
            return [self._seconds2utc(s, reftime=reftime) for s in seconds]
        if isinstance(seconds, UTC) or reftime is None or seconds is None:
            return seconds
        if not isinstance(reftime, UTC):
            reftime = self.stats[reftime]
        return reftime + seconds

    def write(self, filename, format, **kwargs):
        """
        Save current trace into a file  including format specific headers.

        See `Trace.write() <obspy.core.trace.Trace.write>` in ObsPy.
        """
        RFStream([self]).write(filename, format, **kwargs)


def obj2stats(event=None, station=None):
    """
    Map event and station object to stats with attributes.

    :param event: ObsPy `~obspy.core.event.event.Event` object
    :param station: station object with attributes latitude, longitude and
        elevation
    :return: ``stats`` object with station and event attributes
    """
    stats = AttribDict({})
    if event is not None:
        for key, getter in _EVENT_GETTER:
            stats[key] = getter(event)
    if station is not None:
        for key, getter in _STATION_GETTER:
            stats[key] = getter(station)
    return stats


def rfstats(obj=None, event=None, station=None,
            phase='P', dist_range='default', tt_model='iasp91',
            pp_depth=None, pp_phase=None, model='iasp91'):
    """
    Calculate ray specific values like slowness for given event and station.

    :param obj: `~obspy.core.trace.Stats` object with event and/or station
        attributes. Can be None if both event and station are given.
        It is possible to specify a stream object, too. Then, rfstats will be
        called for each Trace.stats object and traces outside dist_range will
        be discarded.
    :param event: ObsPy `~obspy.core.event.event.Event` object
    :param station: dictionary like object with items latitude, longitude and
        elevation
    :param phase: string with phase. Usually this will be 'P' or
        'S' for P and S receiver functions, respectively.
    :type dist_range: tuple of length 2
    :param dist_range: if epicentral of event is not in this intervall, None
        is returned by this function,\n
        if phase == 'P' defaults to (30, 90),\n
        if phase == 'S' defaults to (50, 85)
    :param tt_model: model for travel time calculation.
        (see the `obspy.taup` module, default: iasp91)
    :param pp_depth: Depth for piercing point calculation
        (in km, default: None -> No calculation)
    :param pp_phase: Phase for pp calculation (default: 'S' for P-receiver
        function and 'P' for S-receiver function)
    :param model: Path to model file for pp calculation
        (see `.SimpleModel`, default: iasp91)
    :return: `~obspy.core.trace.Stats` object with event and station
        attributes, distance, back_azimuth, inclination, onset and
        slowness or None if epicentral distance is not in the given interval.
        Stream instance if stream was specified instead of stats.
    """
    if isinstance(obj, (Stream, RFStream)):
        stream = obj
        kwargs = {'event': event, 'station': station,
                  'phase': phase, 'dist_range': dist_range,
                  'tt_model': tt_model, 'pp_depth': pp_depth,
                  'pp_phase': pp_phase, 'model': model}
        traces = []
        for tr in stream:
            if rfstats(tr.stats, **kwargs) is not None:
                traces.append(tr)
        stream.traces = traces
        return stream
    if dist_range == 'default' and phase.upper() in 'PS':
        dist_range = (30, 90) if phase.upper() == 'P' else (50, 85)
    elif dist_range == 'default':
        raise ValueError('Please specify dist_range parameter')
    stats = AttribDict({}) if obj is None else obj
    if event is not None and station is not None:
        stats.update(obj2stats(event=event, station=station))
    dist, baz, _ = gps2dist_azimuth(stats.station_latitude,
                                    stats.station_longitude,
                                    stats.event_latitude,
                                    stats.event_longitude)
    dist = dist / 1000 / DEG2KM
    if dist_range and not dist_range[0] <= dist <= dist_range[1]:
        return
    tt_model = TauPyModel(model=tt_model)
    arrivals = tt_model.get_travel_times(stats.event_depth, dist, (phase,))
    if len(arrivals) == 0:
        raise Exception('TauPy does not return phase %s at distance %s' %
                        (phase, dist))
    if len(arrivals) > 1:
        msg = ('TauPy returns more than one arrival for phase %s at '
               'distance %s -> take first arrival')
        warnings.warn(msg % (phase, dist))
    arrival = arrivals[0]
    onset = stats.event_time + arrival.time
    inc = arrival.incident_angle
    slowness = arrival.ray_param_sec_degree
    stats.update({'distance': dist, 'back_azimuth': baz, 'inclination': inc,
                  'onset': onset, 'slowness': slowness, 'phase': phase})
    if pp_depth is not None:
        model = load_model(model)
        if pp_phase is None:
            pp_phase = 'S' if phase.upper().endswith('P') else 'P'
        model.ppoint(stats, pp_depth, phase=pp_phase)
    return stats



# Copyright 2013-2016 Tom Eulenfeld, MIT license
import contextlib
import os
import shutil
import sys
import tempfile


class _Devnull(object):

    def write(self, _):
        pass


@contextlib.contextmanager
def quiet():
    stdout_save = sys.stdout
    sys.stdout = _Devnull()
    try:
        yield
    finally:
        sys.stdout = stdout_save


@contextlib.contextmanager
def tempdir(delete=True, change_dir=True):
    if delete:
        tempdir = tempfile.mkdtemp(prefix='rf_test')
    else:
        tempdir = os.path.join(tempfile.gettempdir(), 'rf_test_permanent')
        if not os.path.exists(tempdir):
            os.mkdir(tempdir)
    if change_dir:
        cwd = os.getcwd()
        os.chdir(tempdir)
    try:
        yield tempdir
    finally:
        if change_dir:
            os.chdir(cwd)
        if delete and os.path.exists(tempdir):
            shutil.rmtree(tempdir)



# Copyright 2013-2019 Tom Eulenfeld, MIT license
"""
All objects and functions below are after the Tom Eulenfeld's
rf project License below.

The MIT License (MIT)

Copyright (c) 2013-2019 Tom Eulenfeld

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
"""
Functions for receiver function profile calculation.
"""
import numpy as np
from rfsed.rf.util import _add_processing_info, direct_geodetic


_LARGE_BOX_WIDTH = 2000


def _get_box(latlon0, azimuth, length, width=_LARGE_BOX_WIDTH, offset=0):
    """Create a single box."""
    from shapely.geometry import Polygon
    start = direct_geodetic(latlon0, azimuth, offset)
    azis = ((azimuth - 90) % 360, azimuth,
            (azimuth + 90) % 360, (azimuth + 180) % 360)
    dists = (width/2, length, width, length)
    latlon = start
    corners = []
    for a, d in zip(azis, dists):
        latlon = direct_geodetic(latlon, a, d)
        corners.append(latlon[::-1])
    box = {'poly': Polygon(corners),
           'length': length,
           'pos': offset + length/2,
           'latlon': direct_geodetic(start, azimuth, length/2)}
    return box


def get_profile_boxes(latlon0, azimuth, bins, width=_LARGE_BOX_WIDTH):
    """
    Create 2D boxes for usage in `profile()` function.

    :param tuple latlon0: coordinates of starting point of profile
    :param azimuth: azimuth of profile direction
    :param tuple bins: Edges of the distance bins in km (e.g. (0, 10, 20, 30))
    :param width: width of the boxes in km (default: large)
    :return: List of box dicts. Each box has the entries
        'poly' (shapely polygon with lonlat corners), 'length' (length in km),
        'pos' (midpoint of box in km from starting coordinates),
        'latlon' (midpoint of box as coordinates)
    """
    boxes = []
    for i in range(len(bins)-1):
        length = bins[i+1] - bins[i]
        box = _get_box(latlon0, azimuth, length, width, offset=bins[i])
        if i == 0:
            box['profile'] = {}
            box['profile']['latlon'] = latlon0
            box['profile']['azimuth'] = azimuth
            box['profile']['length'] = bins[-1] - bins[0]
            box['profile']['width'] = width
        boxes.append(box)
    return boxes


def _find_box(latlon, boxes, crs=None):
    """Return the box which encloses the coordinates."""
    import cartopy.crs as ccrs
    from shapely.geometry import Point
    if crs is None:
        latlons = [boxes[len(boxes)//2]['latlon']]
        latlon0 = np.median(latlons, axis=0)
        crs = ccrs.AzimuthalEquidistant(*latlon0[::-1])
    pc = ccrs.PlateCarree()
    p = crs.project_geometry(Point(*latlon[::-1]), pc)
    for box in boxes:
        poly = crs.project_geometry(box['poly'], pc)
        if p.within(poly):
            return box


@_add_processing_info
def profile(stream, boxes, crs=None):
    """
    Stack traces in stream by piercing point coordinates in defined boxes.

    :param stream: stream with pre-calculated piercing point coordinates
    :param boxes: boxes created with `get_profile_boxes()`
    :param crs: cartopy projection (default: AzimuthalEquidistant)
    :return: profile stream
    """
    stack = {}
    for tr in stream:
        ppoint = (tr.stats.pp_latitude, tr.stats.pp_longitude)
        box = _find_box(ppoint, boxes, crs=crs)
        if box is None:
            continue
        pos = box['pos']
        comp = tr.stats.channel[-1]
        key = (pos, comp)
        if key not in stack:
            header = {'box_pos': pos,
                      'box_length': box['length'],
                      'box_latitude': box['latlon'][0],
                      'box_longitude': box['latlon'][1],
                      'profile_latitude': boxes[0]['profile']['latlon'][0],
                      'profile_longitude': boxes[0]['profile']['latlon'][1],
                      'profile_azimuth': boxes[0]['profile']['azimuth'],
                      'profile_length': boxes[0]['profile']['length'],
                      'num': 1,
                      'sampling_rate': tr.stats.sampling_rate,
                      'channel': '??' + comp}
            for entry in ('slowness', 'phase', 'moveout', 'processing'):
                if entry in tr.stats:
                    header[entry] = tr.stats[entry]
            stack[key] = tr2 = tr.__class__(data=tr.data, header=header)
            if 'onset' in tr.stats:
                onset = tr.stats.onset - tr.stats.starttime
                tr2.stats.onset = tr2.stats.starttime + onset
        else:
            tr2 = stack[key]
            tr2.data = tr2.data + tr.data
            tr2.stats.num += 1
    for tr2 in stack.values():
        tr2.data = tr2.data / tr2.stats.num
    if hasattr(stream, 'iterable'):  # support tqdm objects
        cls = stream.iterable.__class__
    else:
        cls = stream.__class__
    try:
        profile = cls(traces=stack.values())
    except TypeError:  # stream can be an iterator
        from rf import RFStream
        profile = RFStream(traces=stack.values())
    profile.sort(['channel', 'box_pos'])
    profile.type = 'profile'
    return profile
